require "edx"
require "toolset"
require "toolset_manager"
require "lsp"
require "project"

edx:active("output");

function show_build_result_notification(success)
	if success then
		show_notify(LANG("notify/build/success"));
	else
		show_notify(LANG("notify/build/failed"));
	end;
end;

function show_install_result_notification(success)
	if success then
		show_notify(LANG("notify/build/installed"));
	else
		show_notify(LANG("notify/build/install_failed"));
	end;
end;

function toolchain_update_include_path(all_doc)
	local current_toolset = menu_bar.get_cmake_toolset(true);
	if current_toolset ~= nil and current_toolset.include ~= nil then
		local unmap_path = current_toolset.unmap_path or function(path) return path; end;
		local include_path = {};
		for idx, path in ipairs(current_toolset.include) do
			table.insert(include_path, unmap_path(path));
		end;
		local target_include_path = table.concat(include_path, ";");
		for idx, doc in ipairs(all_doc) do
			doc.include_path = target_include_path;
		end;
	end;
end;

function toolchain_update_include_path_for_opened_documents()
	local all_doc = {mgr.get_all_documents()};
	toolchain_update_include_path(all_doc);
end;

function cmake_update_include_path_for_opened_documents()
	local all_doc = {mgr.get_all_documents()};
	local def_target = menu_bar.get_cmake_default_target();
	if def_target ~= nil and def_target.include ~= nil then
		local target_include_path = table.concat(def_target.include, ";");
		for idx, doc in ipairs(all_doc) do
			doc.include_path = target_include_path;
		end;
	end;
end;

function build_toolset_params(cmd)
	local toolset = cmd.toolset;
	local param_path = nil;
	local path = table.iclone(toolset.bin, toolset.lib);
	if not (toolset.remote or toolset.wsl) then
		local cmake_path = (toolset.tools and toolset.tools["cmake"]) or CMAKE_PATHS[1];
		local gdb_path = (toolset.tools and toolset.tools["gdb"]) or GDB_PATHS[1];
		local ninja_path = (toolset.tools and toolset.tools["ninja"]) or NINJA_PATHS[1];
		local ccache_path = (toolset.tools and toolset.tools["ccache"]) or CCACHE_PATHS[1];
		local xmake_path = (toolset.tools and toolset.tools["xmake"]) or XMAKE_PATHS[1];
		local vcpkg_path = VCPKG_PATHS[1];
		local adapters = (toolset.debug_adapter and toolset.debug_adapter.adapters) or {};
		if cmake_path then
			table.insert(path, make_path(cmake_path, ".."));
		end;
		if gdb_path then
			table.insert(path, make_path(gdb_path, ".."));
		end;
		if ninja_path then
			table.insert(path, make_path(ninja_path, ".."));
		end;
		if ccache_path then
			table.insert(path, make_path(ccache_path, ".."));
		end;
		if xmake_path then
			table.insert(path, make_path(xmake_path, ".."));
		end;
		if vcpkg_path then
			table.insert(path, make_path(vcpkg_path, ".."));
		end;
		for key, info in pairs(adapters) do
			if info.bin then
				table.insert(path, make_path(info.bin, ".."));
			end;
		end;
	end;
	path = table.unique(path);
	param_path = table.concat(table.unique(table.iclone(path, {split_string((os.getenv("PATH") or ""), ";")})), ";");
	local params = {
		[1] = cmd.project_path;
		["PATH"] = param_path;
		["INCLUDE"] = table.concat(toolset.include, ";");
		["LIB"] = table.concat(toolset.lib, ";");
	};
	if toolset.envs then
		for name, val in pairs(toolset.envs) do
			params[name] = val;
		end;
	end;
	return params;
end;

function open_terminal(open_path,skip_current_file)
	local cmd = project:find_build_command(nil, skip_current_file);
	local full_wt, wt_root, wt = find_windows_terminal();
	local cmd_line = "start cmd.exe /K pushd ".."\""..open_path.."\"";
	local working_dir = os.getenv("USERPROFILE") or os.getenv("SystemRoot") or [[c:\]];
	if wt ~= nil then
		-- !!! merge new terminal into exists WindowsTerminal frame as a new tabpage, environment will lost
		-- cmd_line = [["]]..wt..[[ -w 0 cmd.exe /K pushd """]]..open_path..[[""""]];
		cmd_line = [["]]..wt..[[ cmd.exe /K pushd """]]..open_path..[[""""]];
		working_dir = wt_root;
	end;
	if cmd ~= nil and (cmd.type == "cmake" or cmd.type == "xmake") then
		local params = build_toolset_params(cmd);
		params[1] = working_dir;
		edx.shell2(cmd_line, "", params);
	else
		edx.shell2(cmd_line, "", working_dir);
	end
end;

-- 打开工程中的文件
function xws_mgr:update_targets(targets, reload)
	if targets == nil then
		return;
	end;
	local build_targets = {};
	self.__old_targets = self.__old_targets or {};
	local old_targets = self.__old_targets;
	local old_active = self.__old_active;
	local new_targets = {};
	local current_target = menu_bar.get_cmake_default_target();
	current_target = (current_target and current_target.name) or "";
	local cmd = project:find_build_command(nil, true);
	local unmap_path = (cmd and cmd.toolset and cmd.toolset.unmap_path) or function(path) return path; end;

	for idx, item in ipairs(targets) do
		local name = item["name"];
		local target = {
			["name"] = name;
			["output"] = unmap_path(item["output"] or "");
			["type"] = item["type"];
			["status"] = 0;
		};
		if name == current_target then
			target.status = target.status|0x02;
			self.__old_active = target;
			if old_active and old_active.name == name then
				old_active = nil;
			end;
		end;
		table.insert(build_targets, target);
		old_targets[name] = nil;
		new_targets[name] = true;
	end;
	self.__old_targets = new_targets;
	if not reload and old_active ~= nil then
		old_active.status = 0;
		table.insert(build_targets, old_active);
	end;
	self:add_build_targets(build_targets);
	if reload then
		old_targets = table.keys(old_targets);
		if #old_targets > 0 then
			self:remove_build_targets(old_targets);
		end;
	end;
end;

function xws_mgr.on_open_file(self, id, file_path, mode)
	if mgr:open_doc({mode, file_path}) then
		edx:active("editor");
	end;
	return true;
end;

function xws_mgr.on_rename_file(self, id, phase, from_path, to_path)
	-- phase
	-- 0 begin to rename, return true to do rename
	-- 1 file renamed
	-- 2 failed to rename
	local all_doc = {mgr.get_all_documents()};
	if phase == 0 then
		-- stop monitor document
		for idx, doc in ipairs(all_doc) do
			local file_name = doc.file_name;
			if is_file_in_path(file_name, from_path) then
				mgr.monitor_doc(file_name, false);
			end;
		end;
		return true;
	end;
	if phase == 1 then
		project:on_file_renamed(from_path, to_path);
		-- file named successfully
		for idx, doc in ipairs(all_doc) do
			local file_name = doc.file_name;
			if file_name == from_path then
				doc.rename(to_path);
				-- restart monitor document
				mgr.monitor_doc(to_path, true);
				lsp:did_rename_doc(doc, from_path, to_path);
				break;
			end;
			if is_file_in_path(file_name, from_path) then
				local new_path = make_path(to_path, file_name:sub(#from_path + 1));
				doc.rename(new_path);
				-- restart monitor document
				mgr.monitor_doc(new_path, true);
				lsp:did_rename_doc(doc, file_name, new_path);
			end;
		end;
	else
		for idx, doc in ipairs(all_doc) do
			local file_name = doc.file_name;
			if file_name == from_path then
				doc.rename(to_path);
				-- restart monitor document
				mgr.monitor_doc(from_path, true);
				break;
			end;
			if is_file_in_path(file_name, from_path) then
				-- restart monitor old document
				mgr.monitor_doc(file_name, true);
			end;
		end;
	end;
end;

function xws_mgr.on_load_xws(self, xws_file)
	self.is_loaded = true;
	edx:switch_layout("simple");
    print("load solution " .. xws_file);
	menu_bar.show_project_menu();
	mgr.hide_single_tab = false;
    return 1;
end;

function xws_mgr.on_unload_xws(self, xws_file)
	self.is_loaded = false;
	edx:switch_layout("lite");
	print("unload solution " .. xws_file);
	menu_bar.hide_project_menu();
	mgr.search_path = nil;
	mgr.hide_single_tab = true;
	edx:active("editor");
	return 1;
end;

xws_mgr._pre_load_dir = {
	function(self,dir)
		mgr:clear_doc_diagnostic("", "");
		debug_stop();
	end;

	function(self, dir)
		-- try to create CMakeLists.txt for stm32cube mx project
		local cmake_lists = make_path(dir, "CMakeLists.txt");
		if utils:file_time(cmake_lists) then
			-- CMakeLists.txt exists
			return;
		end;
		local cubemx_proj = make_path(dir, ".mxproject");
		if utils:file_time(cubemx_proj) == nil then
			-- not a cubemx project
			return;
		end;
		local cubemx_conf = utils:list_dir(make_path(dir, "*.ioc"));
		local conf_file = io.u8_open(make_path(dir, cubemx_conf), "r");
		if not conf_file then
			print("Can not open ", cubemx_conf);
			return;
		end;
		local target_device = "";
		for line in conf_file:lines() do
			target_device = line:match("^Mcu%.CPN%s*=%s*(STM32......).*");
			if target_device then
				target_device = target_device:lower();
				break;
			end;
		end;
		conf_file:close();
		local proj_name = get_file_name(dir) or "";
		proj_name = proj_name:gsub("%s+", "-"):gsub("[#${}!()]","");
		write_file(make_path(dir, "CMakeLists.txt"),
			make_cmake_list{
				["type"] = "exec";
				["project"] = proj_name;
				["name"] = proj_name;
				["nodep"] = true;
				["embeded"] = true;
				["extra"] = "";
				["target_link_libs"] = [[${STM32CUBE_SDK}]];
				["target_device"] = target_device;
			}
		);
		local toolset = stm32cube_find_toolset_for_device(target_device);
		write_file(make_path(dir, "settings.lua"),
			gen_default_config("Core/Src/main.c", "Debug", toolset, target_device)
		);
	end;
};

xws_mgr._pre_load_cmake_project = {
	function(self, dir, cmake_lists, cmake_lists_data)
		-- try to create settings.lua for pico project
		local settings_lua = make_path(dir, "settings.lua");
		local toolset = pico_find_toolset_for_device("rp2040");
		if toolset == nil or utils:file_time(settings_lua) ~= nil then
			-- no CMakeLists.txt or settings.lua exists
			return false;
		end;
		local is_pico_project = cmake_lists_data:match("^%s*pico_sdk_init%(%s*%).*") ~= nil;
		if not is_pico_project then
			return false;
		end;
		write_file(make_path(dir, "settings.lua"),
			gen_default_config("", "Debug", toolset, "")
		);
		return true;
	end;
};

xws_mgr._post_load_dir = {
	function(self, dir)
		local user_config = make_path(dir, ".edx/user-config.lua");
		if utils:file_time(user_config) then
			-- {CWD}/.edx/user-config.lua
			pinvoke(dofile, user_config);
		end;
		return false;
	end;
	function(self, dir)
		local toolset = menu_bar.get_cmake_toolset(true);
		if toolset.debug_adapter then
			local debug_adapter = toolset.debug_adapter;
			local config_adapter = project:get_extra_setting("debug_adapter", "");
			menu_bar.set_cmake_debug_adapter(config_adapter);
			debug_adapter:switch_adapter(menu_bar.get_cmake_debug_adapter());
			menu_bar.update_cmake_debug_adapter_transport(debug_adapter.transports);
			menu_bar.update_cmake_debug_adapter_speed(debug_adapter.speeds);
			if #debug_adapter.transports == 0 then
				menu_bar.hide_cmake_items("debug_adapter_transport");
			else
				menu_bar.show_cmake_items("debug_adapter_transport");
			end;
			if #debug_adapter.speeds == 0 then
				menu_bar.hide_cmake_items("debug_adapter_speed");
			else
				menu_bar.show_cmake_items("debug_adapter_speed");
			end;
			menu_bar.set_cmake_debug_adapter_speed(project:get_extra_setting("debug_adapter_speed", ""));
			menu_bar.set_cmake_debug_adapter_transport(project:get_extra_setting("debug_adapter_transport", ""));
		end;
		return false;
	end;
};

xws_mgr._pre_unload_dir = {
	function(self, dir)
		local toolset = menu_bar.get_cmake_toolset(true);
		if toolset.debug_adapter then
			local adapter_info = toolset.debug_adapter.adapters[menu_bar.get_cmake_debug_adapter()];
			project:set_extra_setting("debug_adapter", menu_bar.get_cmake_debug_adapter());
			if #adapter_info.speeds > 0 then
				project:set_extra_setting("debug_adapter_speed", menu_bar.get_cmake_debug_adapter_speed());
			else
				project:set_extra_setting("debug_adapter_speed", nil);
			end;
			if #adapter_info.transports > 0 then
				project:set_extra_setting("debug_adapter_transport", menu_bar.get_cmake_debug_adapter_transport());
			else
				project:set_extra_setting("debug_adapter_transport", nil);
			end;
		end;
		project:set_extra_setting("com_baud", menu_bar.get_cmake_serial_baud());
		local mon_port, mon_baud = edx:get_serial_monitor_info();
		project:set_extra_setting("mon_baud", mon_baud);
		return false;
	end;
};

xws_mgr._post_unload_dir = {
	function(self, dir)
		self.__old_targets = nil;
		self.__old_active = nil;
		edx:add_recently_folder(dir);
		mgr:clear_doc_diagnostic("", "");
		menu_bar.set_cmake_default_target(nil);
		debug_stop();
		return false;
	end;
};

function xws_mgr._do_load_dir(self, dir)
	if dir:sub(-1) == "\\" or dir:sub(-1) == "/" then
		dir = dir:sub(1,-2);
	end;
	if self._pre_load_dir then
		for idx, cb in ipairs(self._pre_load_dir) do
			if pinvoke(cb, self, dir) then
				break;
			end;
		end;
	end;
	
	if self._pre_load_cmake_project then
		local cmake_lists = make_path(dir, "CMakeLists.txt");
		if utils:file_time(cmake_lists) then
			local cmake_lists_data = load_file(cmake_lists);
			for idx, cb in ipairs(self._pre_load_cmake_project) do
				if pinvoke(cb, self, dir, cmake_lists, cmake_lists_data) then
					break;
				end;
			end;
		end;
	end;
	
	if not project:load(dir) then
		lsp:_try_init_lsp(dir);
		menu_bar.show_project_menu();
	end;
	if self._post_load_dir then
		for idx, cb in ipairs(self._post_load_dir) do
			if pinvoke(cb, self, dir) then
				break;
			end;
		end;
	end;
end;

function xws_mgr.on_load_dir(self, dir)
	self.is_loaded = true;
	print("load dir " .. dir);
	dbg_call(function()
			self:_do_load_dir(dir);
	end,"Load dir error!");
	local code = edx.load_code_version_service(dir);
	edx:switch_layout("simple");
	if code ~= 0 and code ~= -1 then
		local error_codes = {
			[-36] = "notify/revision/error/invalid_owner";
		};
		show_notify(string.format(LANG(error_codes[code] or "notify/revision/error/can_not_error"), code, dir));
	end;
	mgr.search_path = dir;
	mgr.hide_single_tab = false;
	edx.default_save_path = dir;
	return 1;
end;

function xws_mgr.on_unload_dir(self, dir)
	if self._pre_unload_dir then
		for idx, cb in ipairs(self._pre_unload_dir) do
			if pinvoke(cb, self, dir) then
				break;
			end;
		end;
	end;
	
	self.is_loaded = false;
	print("unload dir " .. dir);
	edx:switch_layout("lite");
	edx.unload_code_version_service();
	menu_bar:hide_project_menu();
	menu_bar:hide_cmake_menu();
	menu_bar:hide_debug_menu();
	project:unload();
	lsp:terminate();
	mgr.search_path = nil;
	mgr.hide_single_tab = true;
	edx.default_save_path = "";
	edx:active("editor");

	if self._post_unload_dir then
		for idx, cb in ipairs(self._post_unload_dir) do
			if pinvoke(cb, self, dir) then
				break;
			end;
		end;
	end;
	return 1;
end;

function xws_mgr.on_delete_file(self, id, file_path)
    -- 删除文件之前强制关闭这个文档
	mgr:remove_doc(file_path);
	project:on_file_removed(file_path);
end;

function xws_mgr.on_cmd(self, id)
	local cmd = self._cmd_map[id];
	if cmd then
		return pinvoke(cmd, self, id);
	end;
	return 0;
end;

xws_mgr._cmd_map = {
	[_op.git_init] = function(self, id)
		local git_ignore_file = project:make_path(".gitignore");
		local git_ignore_content = ".edx/\nsettings.lua\n";
		if utils:file_time(git_ignore_file) then
			local old_file = load_file(git_ignore_file);
			if old_file:sub(#old_file) ~= "\n" then
				git_ignore_content = old_file .. "\n# EDX files\n" .. git_ignore_content;
			else
				git_ignore_content = old_file .. "# EDX files\n" .. git_ignore_content;
			end;
		end;
		write_file(git_ignore_file, git_ignore_content);
		return 0;
	end;
	[_op.git_diff_file] = function(self, id)
		local item_path, item_type = xws_mgr.get_selected_item();
		if item_type == 4 or item_type == 3 or item_type == 2 then
			edx.cvs_diff_file(item_path);
		end;
		return true;
	end;
	[_op.xws_project_configuration] = function(self, id)
		local item_path, item_type, item_id = xws_mgr.get_context_menu_item();
		if item_type == 0x11 then
			config_debug_run:show_target(xws_mgr.get_item_name(item_id));
		else
			config_debug_run:show();
		end;
		return true;
	end;
	[_op.project_configuration] = function(self, id)
		config_debug_run:show();
		return true;
	end;
	[_op.xws_pop_menu] = function(self, id)
		if not self.is_xws_loaded() then
			show_notify(LANG("notify/error/solution_not_opened"));
			return 1;
		end;
		local menu = self.popup_menu;
		local item_path, item_type, item_id = xws_mgr.get_context_menu_item();
		local open_path = item_path or edx.startup_path;
		local item_status = xws_mgr.get_item_status(item_id);
		if item_type == 0x10 or item_type == 0x11 then
			menu:show_item(_op.xws_open_file, false, false);
			menu:show_item(1, false, true); -- new
			menu:show_item(_op.xws_fold_sub_items, false, false);
			menu:show_item(_op.xws_unfold_sub_items, false, false);
			menu:show_item(_op.xws_close_opend_file, false, false);
			menu:show_item(_op.git_init, false, false);
			menu:show_item(_op.xws_remove_selected_item, false, false);
			menu:show_item(_op.xws_cut_selected_item, false, false);
			menu:show_item(_op.xws_paste_file, false, false);
			menu:show_item(_op.xws_rename_selected_item, false, false);
			if item_type == 0x11 then
				local current_target = menu_bar.get_cmake_default_target();
				current_target = (current_target and current_target.name) or "";
				menu:show_item(_op.xws_build_target, true, false);
				menu:show_item(_op.xws_rebuild_target, true, false);
				menu:show_item(_op.debug_start, true, false);
				menu:show_item(_op.build_and_run, true, false);
				menu:enable_item(_op.build_target, not is_debugger_running(), false);
				menu:enable_item(_op.rebuild_target, not is_debugger_running(), false);
				menu:enable_item(_op.debug_start, not is_debugger_running(), false);
				menu:show_item(_op.xws_set_startup_target, true, false);
			else
				menu:show_item(_op.xws_build_target, false, false);
				menu:show_item(_op.xws_rebuild_target, false, false);
				menu:show_item(_op.build_and_run, false, false);
				menu:show_item(_op.debug_start, false, false);
				menu:show_item(_op.xws_set_startup_target, false, false);
			end;
		else
			menu:show_item(1, true, true); -- new
			menu:show_item(_op.xws_build_target, false, false);
			menu:show_item(_op.xws_rebuild_target, false, false);
			menu:show_item(_op.build_and_run, false, false);
			menu:show_item(_op.debug_start, false, false);
			menu:show_item(_op.xws_set_startup_target, false, false);
			menu:show_item(_op.xws_close_opend_file, true, false);
			menu:show_item(_op.git_init, true, false);
			menu:show_item(_op.xws_remove_selected_item, true, false);
			menu:show_item(_op.xws_cut_selected_item, true, false);
			menu:show_item(_op.xws_paste_file, true, false);
			menu:show_item(_op.xws_rename_selected_item, true, false);
			if item_type == 4 then
				-- file item
				menu:show_item(_op.xws_open_file, true, false);
				menu:show_item(_op.xws_fold_sub_items, false, false);
				menu:show_item(_op.xws_unfold_sub_items, false, false);
			else
				menu:show_item(_op.xws_open_file, false, false);
				local item_unfolded = (item_status&0x02) == 0x00;
				-- has folded sub items
				local can_unfold = item_unfolded and (item_status&0x08) == 0x08;
				menu:show_item(_op.xws_unfold_sub_items, can_unfold, false);
				-- has unfolded sub items
				local can_fold = item_unfolded and (item_status&0x04) == 0x04;
				menu:show_item(_op.xws_fold_sub_items, can_fold, false);
			end;
			if item_type == 2 or item_type == 1 then
				menu:enable_item(_op.xws_remove_selected_item, false, false);
				menu:enable_item(_op.xws_cut_selected_item, false, false);
				menu:enable_item(_op.xws_rename_selected_item, false, false);
			else
				menu:enable_item(_op.xws_remove_selected_item, true, false);
				menu:enable_item(_op.xws_cut_selected_item, true, false);
				menu:enable_item(_op.xws_rename_selected_item, true, false);
			end;
		end;

		local git_menu = menu.sub_menu(_op.git_init, false);
		if git_menu then
			local is_cvs_ready = (edx:cvs_ready() == 0);
			git_menu:show_item(_op.git_init, not is_cvs_ready, false);
			git_menu:show_item(1, not is_cvs_ready, true);
			git_menu:enable_item(_op.git_diff_file, is_cvs_ready, false);
			git_menu:enable_item(_op.git_commit, is_cvs_ready, false);
			git_menu:enable_item(_op.git_pull, is_cvs_ready, false);
			git_menu:enable_item(_op.git_push, is_cvs_ready, false);
			git_menu:enable_item(_op.git_checkout, is_cvs_ready, false);
			git_menu:enable_item(_op.git_fetch, is_cvs_ready, false);
			git_menu:enable_item(_op.ext_git_open_history, is_cvs_ready, false);
			git_menu:enable_item(_op.ext_git_open_gui, is_cvs_ready, false);
		end;
		return 0;
	end;
	[_op.xws_build_target] = function(self, id)
		return global_cmd_map[id](edx, id);
	end;
};

function xws_mgr.on_create_file(self, id, file_path)
	local workspace = self.workspace;
	local is_cmake_file = ("CMakeLists.txt" == file_path:gsub("(.*[/\\])", ""));
	if is_cmake_file then
		local file_dir = make_path(file_path, "..");
		local folder_name = file_dir:match([[.*[/\]([^/\]+)]]);
		if not folder_name then
			folder_name = "NONAME";
		end;
		local file_content = make_cmake_list{
			["type"] = "exec";
			["project"] = folder_name;
			["name"] = folder_name;
			["nodep"] = false;
			["embeded"] = false;
			["extra"] = "# add_link_options(<LINKER_FLAGS>)";
		};
		if mgr:open_doc(file_path) then
			local doc = mgr:get_document(file_path);
			doc:set_text(file_content);
			doc:save();
			doc:set_cursor(0,0);
		end;
		local root_cmake_path = make_path(workspace, "CMakeLists.txt");
		if file_path == root_cmake_path then
			-- 转换工程为CMake工程
			menu_bar:hide_project_menu();
			self:_do_load_dir(workspace);
		end;
	end;
	if project:is_ready() then
		if project:make_path("vcpkg.json") == file_path and mgr:open_doc(file_path) then
			local doc = mgr:get_document(file_path);
			doc:set_text("{\n\t\"dependencies\": [\n\t\"\"\n\t]\n}");
			doc:save();
			doc:set_cursor(2,2);
		else
			project:on_file_created(file_path);
		end;
	end;
end;

function xws_mgr.on_create_dir(self, id, dir)
end;

-- 更新最近使用的文件列表
function edx.__strip_recently(self, recently)
	local name_map = {};
	local new_recently = {};
	if recently == nil then
		return new_recently;
	end;
	for idx, val in ipairs(recently) do
		if string.sub(val, 1, 2) ~= [[\\]] then
			val = self:regular_path(val) or val;
		end
		if not name_map[val] and utils:file_time(val) ~= nil then
			name_map[val] = true;
			table.insert(new_recently, val);
		end;
	end;
	return new_recently;
end;
function edx.reload_recently_menu(self)
	local env_time = utils:file_time(edx_env_path);
	if env_time <= self.__env_time then
		return false;
	end;
	local rf = io.u8_open(edx_env_path);
	if rf ~= nil then
		self.__env_time = utils:file_time(edx_env_path);
		local new_env = utils.fromjson(rf:read()) or {};
		self.env.recently = edx:__strip_recently(new_env.recently);
		self.env.recently_folders = edx:__strip_recently(new_env.recently_folders);
		rf:close();
		return true;
	end;
	return false;
end;
function edx.update_recently_menu(self)
    local recently_docs = self.env.recently;
	local recently_folders = self.env.recently_folders;
    local file_menu = menu_bar.sub_menu(0x8000001, false);
	file_menu.remove_item(0x8100001, false);
	file_menu.remove_item(0x8100002, false);
	file_menu.remove_item(0x8200001, false);
	if sizeof(recently_docs) ~= 0 or sizeof(recently_folders) ~= 0 then
		file_menu.insert_line(0x8000000, 0x8200001, false);
	end;
    if sizeof(recently_docs) ~= 0 then
		file_menu.insert_item(0x8000000, LANG("menu_bar/file/recently_files"), 0x8100001, false);
        local recently_menu = file_menu.sub_menu(0x8100001, false);
		for idx, name in pairs(recently_docs) do
			local item_id = recently_menu.insert_item(-1, name, RECENTLY_ITEM_ID_BASE + tonumber(idx), true);
			recently_menu.set_accept_mbtn(item_id, true, false);
        end;
    end ;
	if sizeof(recently_folders) ~= 0 then
		file_menu.insert_item(0x8000000, LANG("menu_bar/file/recently_folders"), 0x8100002, false);
		local recently_menu = file_menu.sub_menu(0x8100002, false);
		for idx, name in pairs(recently_folders) do
			local item_id = recently_menu.insert_item(-1, name, RECENTLY_FOLDER_ID_BASE + tonumber(idx), true);
			recently_menu.set_accept_mbtn(item_id, true, false);
        end;
    end;
end;
function edx.add_recently_doc(self, doc_name, is_remove)
	if string.sub(doc_name, 1, 2) ~= [[\\]] then
		local doc_path = self:regular_path(doc_name);
		if doc_path == nil or doc_path == "" then
			-- not a valid local document
			return;
		end;
	end
	self:reload_recently_menu();
    if doc_name == "" or doc_name == nil then
        return ;
    end ;
	local new_recently_docs = {};
	local counter = 0;
	if not is_remove then
		counter = 1;
		new_recently_docs[counter] = doc_name;
	end;
    local recently_docs = self.env.recently;
    self.env.recently = new_recently_docs;
    for idx, name in pairs(recently_docs) do
		if name and #name > 0 and name ~= doc_name then
			counter = counter + 1;
			new_recently_docs[counter] = name;
        end;
		if counter >= 16 then
            break;
        end;
	end;
	self:update_recently_menu();
	self:save_env();
end;
function edx.add_recently_folder(self, folder_path, is_remove)
	self:reload_recently_menu();
	if folder_path == "" or folder_path == nil then
        return ;
    end ;
	local new_recently_folders = {};
	local counter = 0;
	if not is_remove then
		counter = 1;
		new_recently_folders[counter] = folder_path;
	end;
	local recently_folders = self.env.recently_folders;
	self.env.recently_folders = new_recently_folders;
	for idx, name in pairs(recently_folders) do
		if name and #name > 0 and name ~= folder_path then
			counter = counter + 1;
			new_recently_folders[counter] = name;
        end;
		if counter >= 16 then
            break;
        end;
    end;
	self:update_recently_menu();
	self:save_env();
end;
function edx.remove_recently_folder(self, folder_path)
	self:add_recently_folder(folder_path, true);
end;
function edx.remove_recently_doc(self, doc_name)
	self:add_recently_doc(doc_name, true);
end;
-- 加载EDX环境
function edx.load_env(self)
	local rf = io.u8_open(edx_env_path);
	if rf ~= nil then
		self.__env_time = utils:file_time(edx_env_path);
		self.env = utils.fromjson(rf:read("a")) or {layout = DEFAULT_LAYOUT};
		rf:close();
		self.env.layout = table.clone(DEFAULT_LAYOUT, self.env.layout);

		self.env.recently = edx:__strip_recently(self.env.recently);
		self.env.recently_folders = edx:__strip_recently(self.env.recently_folders);
    end;
    -- 更新最近使用的文件列表菜单
    self:update_recently_menu();

    local main = self.env.main;
    if main then
        -- 恢复上一次的窗口大小
        self:set_position(main.position.left, main.position.top, main.position.width, main.position.height);
        if main.show == 2 then
            self.show_status = main.show;
        end ;
    end ;
end;
-- 保存EDX环境
function edx.save_env(self)
    self:switch_layout(nil);
    local position = { self:get_normal_position() };
    self.env.main = {
        position = {
            left = position[1];
            top = position[2];
            width = position[3];
            height = position[4];
        };
        show = self.show_status;
    };
	local rf = io.u8_open(edx_env_path, "w+");
    rf:write(utils.tojson(self.env));
	rf:close();
	self.__env_time = utils:file_time(edx_env_path);
end;

function edx.load_toolset(self)
	local rf = io.u8_open(toolset_env_path);
	if rf ~= nil then
		local custom_toolsets = rf:read("a");
		if custom_toolsets then
			custom_toolsets = utils.fromjson(custom_toolsets);
		end;
		update_custom_toolsets(custom_toolsets);
	end;
end;

function edx.save_toolset(self)
	local custom_toolsets = collect_custom_toolsets();
	local rf = io.u8_open(toolset_env_path, "w+");
	rf:write(utils.tojson(custom_toolsets));
	rf:close();
end;

-- 切换docking布局, layout 可以是 lite, simple, debug, fullscreen, fullscreen_lite
function edx.switch_layout(self, layout)
    if self.last_layout ~= nil then
        -- 保存当前的layout
        self.env.layout[self.last_layout] = self.layout;
    end ;
    self.last_layout = layout;
    if self.env.layout[self.last_layout] == nil then
        return false;
	end ;
	local new_layout = self.env.layout[self.last_layout];
	local pane_name_map = {
		["Watch"] = "watch";
		["Locals"] = "locals";
		["Call Stack"] = "stack";
		["Threads"] = "threads";
		["Modules"] = "modules";
		["Registers"] = "registers";
		["Debugger"] = "debugger";
		["Output"] = "output";
		["Find & Replace"] = "find";
		["Build Results"] = "build_result";
		["LSP"] = "lsp";
		["Solution Explorer"] = "solution";
		["Class View"] = "class";
		["Resource View"] = "resource";
		["Property"] = "properties";
		["Structure View"] = "structure";
		["Revisions"] = "revisions";
	};
	-- fix old layout data
	new_layout = new_layout:gsub([[>([%d%s%w]-)</]],
		function(val)
			local name = pane_name_map[val];
			if name then
				return ">" .. name .. "</";
			end;
			return ">" .. val .. "</";
		end
	);
	self.layout = new_layout;
end;

mgr.completion_dispatcher = {
};

require "cmake_completion"
require "xmake_completion"
require "clangd_completion"
require "lua_completion"
require "java_completion"

mgr.completion_dispatcher = {
	[content_type_cxx] = "on_cxx_completion";
	[content_type_dlang] = "on_cxx_completion";
	[content_type_lua] = "on_lua_completion";
	[content_type_java] = "on_cxx_completion";
	[content_type_cmake] = "on_cmake_completion";
	[content_type_xmake] = "on_xmake_completion";
};

function mgr:save_all_named()
	self:save_all(1);
end;

function mgr:on_completion(status, helper)
	if status == 0 then
		self.__completing = 1;
	elseif status == 1 or status == 3 then
		self.__completing = nil;
	end;
	local doc = self.current_document;
	if status == 3 then
		lsp:cancel_completion();
	end;
    -- print( "complete ", status, " doc type ", doc.content_type );
	local dispatcher = self.completion_dispatcher[doc.content_type];
	dispatcher = self[dispatcher];
	if not dispatcher then
		return 0;
	end;
	-- print("do complete: ", dispatcher);
	return pinvoke(dispatcher, self, doc, status, helper);
end;

function mgr:on_completion_tip(status, helper)
	local doc = mgr.current_document;
	if status == 0 then
		return lsp:find_signature_help(doc, helper);
	end;
	if status == 4 then
		lsp:cancel_signature_help();
	end;
	if status == 2 then
		local txt = "";
		if mgr._last_column ~= nil then
			if mgr._last_column < doc.cursor_column then
				txt = doc.get_text(doc.cursor_line, doc.cursor_column-1, 1);
			elseif mgr._last_column > doc.cursor_column then
				txt = doc.get_text(doc.cursor_line, doc.cursor_column, 1);
			end
		end;
		mgr._last_column = doc.cursor_column;
		if txt == "," or txt == "(" or txt == ")" then
			lsp:find_signature_help(doc, helper);	
		end;
	end;
	return 1;
end;

function mgr:on_locate_file(from, file_name, begin_line, begin_col, end_line, end_col)
	mgr.record_cursor_history();
	if file_name:match([[^\\wsl%$\.*$]]) or utils:file_time(file_name) then
		-- keep the wsl path
	elseif __last_toolset and __last_toolset.unmap_path then
		local unmap_path = __last_toolset.unmap_path;
		file_name = unmap_path(file_name);
	end;
	if mgr.open_doc(file_name) then
		local cur_doc = mgr.current_document;
		cur_doc:set_cursor(begin_line, begin_col);
		cur_doc:select_text(begin_line, begin_col, end_line, end_col);
		edx.do_cmd(_op.center_cursor_line);
	else
		print("failed to locate_file ", file_name, " ", begin_line, ",", begin_col, ",", end_line, ",", end_col);
	end;
end;

function mgr:on_closing(doc)
	local doc_name = doc.file_name;
	if edx.__last_tab_doc and edx.__last_tab_doc.id == doc.id then
		edx.__last_tab_doc = nil;
	end;
	if doc_name == "" or doc_name == nil then
		return ;
	end ;
	edx:add_recently_doc(doc_name);
	if doc.doc_type ~= doc_type.text then
		-- not a text file
		return;
	end;
	lsp:doc_closed(doc);
	project:store_breakpoints(doc)
	project:store_file_states(doc)
end;

local function prepare_opened_doc(doc, restore_states)
	if restore_states == nil then
		restore_states = true;
	end;
	lsp:doc_opened(doc);
	lsp:list_doc_structure(doc, nil); -- preload the document structure

	local def_target = menu_bar.get_cmake_default_target();
	if def_target ~= nil and def_target.include ~= nil then
		doc.include_path = table.concat(def_target.include, ";");
	else
		toolchain_update_include_path({doc});
	end;
	local file_name = doc.file_name;
	if project:is_xmake() and project:make_path("xmake.lua") == file_name then
		doc.content_type = content_type_xmake;
	elseif project:is_esp_idf() and project:make_path("sdkconfig") == file_name then
		doc.content_type = content_type_makefile;
	elseif project:is_bouffalo_sdk() and project:make_path("proj.conf") == file_name then
		doc.content_type = content_type_cmake;
	end;
	if restore_states then
		project:restore_breakpoints(doc);
		project:restore_file_states(doc);
	end;
end;

function mgr:on_opened(doc)
	if doc.doc_type ~= doc_type.text then
		-- not a text file
		return;
	end;
	prepare_opened_doc(doc);
end;

function mgr:on_saved(doc)
	if doc.doc_type ~= doc_type.text then
		-- not a text file
		return;
	end;

	if doc.content_type == content_type_cxx and not project:is_ready() then
		-- project is not ready
		local cmd = project:find_build_command();
		if cmd ~= nil and cmd["type"] == "direct" and cmd.toolset ~= lsp:get_startup_toolset(doc) then
			-- startup toolset changed
			lsp:reset(cmd);
			return;
		end;
	end;
	
	lsp:doc_saved(doc);
	local file_name = doc.file_name;
	if file_name == edx_config_path then
		edx:reload_config();
	elseif project:is_ready() then
		if file_name == project:make_path("settings.lua") then
			project:reload_settings();
		elseif project:is_project_script(file_name) then
			edx:on_cmd(_op.cmake_build_cache);
		end;
	end;
end;

function mgr:on_renamed(doc, old_name, new_name)
	if doc.doc_type ~= doc_type.text then
		-- not a text file
		return;
	end;

	lsp:doc_renamed(doc, old_name, new_name);
	lsp:doc_closed(doc);
	project:on_file_renamed(old_name, new_name);
	prepare_opened_doc(doc, false);
end;

function mgr:on_reloaded(doc, reload_type)
	if doc.doc_type ~= doc_type.text then
		-- not a text file
		return;
	end;
	if reload_type == 1 then
		lsp:doc_closed(doc);
		prepare_opened_doc(doc);
		self:open_doc(doc);
	elseif reload_type == 2 then
		-- reloaded in stream mode
		lsp:doc_closed(doc);
	end;
end;

function mgr:on_saved_as(doc, old_name)
	if doc.doc_type ~= doc_type.text then
		-- not a text file
		return;
	end;

	if not old_name or old_name == "" then
		-- save a new file
		self:on_opened(doc);
		return;
	end;
	lsp:did_rename_doc(doc, old_name, doc.file_name);
	lsp:doc_closed(doc);
	prepare_opened_doc(doc);
end;

function mgr:on_char(c)
	table.insert(self.__last_cmd, 1, c);
	local doc = self.current_document;

    -- C++ 自动补全
	if doc.content_type == content_type_cxx then
		if (c == "\"" or c == "<" or c == ">" or c == "." or c == ":" or c == "#" or c == "(" or c == ")" or c == "," or c == ";") then
            doc.write_char(c);
			if c == ":" then
				local cursor_offset = doc.cursor_column;
				if cursor_offset < 256 then
					local text = doc.get_text(doc.cursor_line, 0, cursor_offset+256);
					if text:match("^%s*public%s*:%s*$") or
						text:match("^%s*private%s*:%s*$") or
						text:match("^%s*protected%s*:%s*$") or
						text:match("^%s*default%s*:%s*$") or
						text:match("^%s*case%s*.*:%s*$") or
						text:match("^%s*[_0-9a-zA-Z]+%s*:+%s*$") then
						edx.do_cmd(_op.format_selection);
						return 1;
					end;
				end;
			elseif c == ")" or c == "," then
				edx.do_cmd(_op.update_completion_tip);
				edx.do_cmd(_op.update_completion_list);
				return 1;
			elseif c == ";" then
				edx.do_cmd(_op.completion_end_tip);
				edx.do_cmd(_op.completion_end_list);
				return 1;
			end;
			local cursor_offset = doc.cursor_column;
			local text;
			if cursor_offset < 256 then
				text = doc.get_text(doc.cursor_line, 0, cursor_offset);
			else
				text = doc.get_text(doc.cursor_line, cursor_offset - 256, cursor_offset);
			end;
			if text ~= nil then
				if cursor_offset < 256 then
					local match_text = text:match("^[ \t]*#[ \t]*include[ \t]+[\"<].*$");
					if match_text ~= nil then
						-- 自动补全include
						if not self.__completing then
							edx.do_cmd(_op.show_completion_list);
						end;
						return 1;
					end;
					local match_text = text:match("^[ \t]*#[ \t]*$");
					if match_text ~= nil then
						-- 自动补全#
						if not self.__completing then
							edx.do_cmd(_op.show_completion_list);
						end;
						return 1;
					end;
				end;
                local match_text = text:match("^.*::$");
                if match_text ~= nil then
                    -- 自动补全xxx::
					-- edx.do_cmd(_op.show_completion_list);
					lsp:do_completion(doc, nil);
					return 1;
                end;
                local match_text = text:match("^.*%-%>$");
                if match_text ~= nil then
                    -- 自动补全xxx->
					-- edx.do_cmd(_op.show_completion_list);
					lsp:do_completion(doc, nil);
					return 1;
                end;
				local match_text = text:match("^.*[a-zA-Z_][a-zA-Z_0-9]*%s*[%(]$");
				if match_text ~= nil then
					-- 自动补全xxx(
					edx.do_cmd(_op.show_completion_tip);
					-- edx.do_cmd(_op.update_completion_list);
					lsp:do_completion(doc, nil);
					return 1;
				end;
				local match_text = text:match("^.*%.%.$");
				if match_text ~= nil then
					-- skip ..
					return 1;
				end;
                local match_text = text:match("^.*%.$");
                if match_text ~= nil then
                    -- 自动补全xxx.
					-- edx.do_cmd(_op.show_completion_list);
					lsp:do_completion(doc, nil);
					return 1;
                end;
            end;
            return 1;
        end;
    end;
    -- LUA自动格式化
    if c == ';' and doc.content_type == content_type_lua then
		doc.write_char(c);
        edx.do_cmd(_op.format_selection);
        return 1;
    end;
    -- XML HTML自动格式化
    if doc.content_type == content_type_html or doc.content_type == content_type_xml then
        if c == '>' then
			doc.write_char(c);
            edx.do_cmd(_op.format_selection);
            return 1;
        end ;
    end;
    return 0;
end;

function mgr:on_deactive(doc)
	lsp:cancel_completion(doc);
	lsp:cancel_signature_help(doc);
	lsp:cancel_list_doc_structure(doc);
end;

function mgr:on_active(doc)
	local cmd = project:find_build_command();
	if cmd and cmd.debug_type then
		menu_bar:show_debug_menu();
	else
		menu_bar:hide_debug_menu();
	end;
	if edx:is_visible("class") then
		-- refresh document structure
		lsp:list_doc_structure(doc, nil);
	end;
end;

function mgr:on_tab_event(event, doc)
	if event == "rbtn_clicked" then
		edx.__last_tab_doc = {
			id = doc.id;
			name = doc.file_name;
		};
		local tab_menu = edx.get_object("menu:doc_tab");
		if tab_menu then
			tab_menu:show_item(_op.locate_tab_file, xws_mgr.is_loaded == true, false);
		end;
	else
		edx.__last_tab_doc = nil;
	end;
	-- print("tab event: ", event, ", ", doc);
	return 0;
end;

function mgr:on_mouse_hover(doc, x, y, line, col)
	local dbg = edx.dbg;
	if dbg and dbg:is_running() then
		dbg:hover_info(doc, line, col, dbg.beautifier);
		return;
	end;
	lsp:hover_info(doc, line, col);
end;

function mgr:on_mouse_wheel(doc, delta, key_status)
	-- 0x0001 lbtn is down
	-- 0x0002 rbtn is down
	-- 0x0004 shift is down
	-- 0x0008 ctrl is down
	-- 0x0010 mbtn is down
	-- 0x0020 xbtn1 is down
	-- 0x0040 xbtn2 is down
	if (key_status&0x0C) == 0x0008 then
		-- ctrl pressed
		if delta > 0 then
			edx:on_cmd(_op.zoom_in, 0);
		else
			edx:on_cmd(_op.zoom_out, 0)
		end;
		return true;
	end;
	return false;
end;

function mgr:on_icon_clicked(doc, line)
	dbg_toggle_breakpoint(doc, line);
end;

function edx:on_exit()
	edx:save_env();
	edx:save_toolset();
	lsp:terminate();
end;

local _edx_popup_menu_handler = {
	[0x8200002] = { -- jump to menu
		[_op.list_doc_structure] = {
			["doc_type"] = doc_type.text;
			["content_type"] = "cpp";
			["op"] = "show";
		};
		-- [id] = function(id, doc, menu, owner_menu, owner_item_id) ... end;
	};
	[0x8000001] = function(id, doc, menu, owner_menu, owner_item_id)
		if edx:reload_recently_menu() then
			edx:update_recently_menu();
		end;
	end;
};

function edx:on_popup_menu(menu, owner_menu, owner_item_id)
	-- print(string.format("owner:%x", owner_item_id));
	if owner_menu then
		-- POPUP SUBMENU
		if 0xA100004 == owner_item_id or 0xA000001 == owner_item_id then
			-- update serial ports
			menu_bar.update_cmake_serial_port();
		end;
		local handlers = _edx_popup_menu_handler[owner_item_id] or {};
		local doc = mgr.current_document;
		if type(handlers) == "function" then
			pinvoke(handlers, id, doc, menu, owner_menu, owner_item_id);
			return;
		end;
		for id, info in pairs(handlers) do
			if type(info) == "function" then
				pinvoke(info, id, doc, menu, owner_menu, owner_item_id);
			else
				local test_ret = (info.doc_type == doc.doc_type and content_type_name[doc.content_type] == info.content_type);
				if info.op == "show" then
					menu.show_item(id, test_ret, false);
				else
				end;
			end;
		end;
	else
		-- POPUP MENU
	end;
end;

edx.command_handlers={
	["output"] = {
		["reset"] = function(self, cmd, target)
			if target == "lsp" then
				print("Reseting LSP...");
				local cmd = project:find_build_command(nil, true);
				lsp:reset(cmd);
				return 1;
			end;
			return 0;
		end;
		["clear"] = function(self, cmd, target)
			if target == nil then
				target = "output";
			end;
			local valid_targets = {
				["output"] = true;
				["find"] = true;
				["lsp"] = true;
				["debugger"] = true;
				["serial_ports"] = true;
				["build_result"] = true;
			};
			if valid_targets[target] == nil then
				print("clear [", table.concat(table.keys(valid_targets), "|"), "]");
				return 1;
			end;
			self:clear(target);
			return 1;
		end;
		["debug"] = function(self, cmd, dbg_type, target, arguments, cwd, envs)
			if dbg_type == nil then
				print("debug [windbg|gdb] {target} [arguments] [working path]");
				print("debug map {local path} {remote path}");
				return 1;
			end;
			if dbg_type == "map" then
				debug_map_path({target, arguments});
				return 1;
			end;
			if dbg_type ~= "windbg" and dbg_type ~= "gdb" then
				envs = cwd;
				cwd = arguments;
				arguments = target;
				target = dbg_type;
				dbg_type = "windbg";
			end;
			debug_program(target, dbg_type, arguments, cwd, envs);
			return 1;
		end;
		["gdb"] = function(self, cmd, target, arguments, cwd, envs)
			if target == nil then
				print("gdb {target} [arguments] [working path]");
				return 1;
			end;
			debug_program(target, "gdb/mi", arguments, cwd, envs);
			return 1;
		end;
		["windbg"] = function(self, cmd, target, arguments, cwd, envs)
			if target == nil then
				print("windbg {target} [arguments] [working path]");
				return 1;
			end;
			debug_program(target, "windbg", arguments, cwd, envs);
			return 1;
		end;
		["syntax"] = function(self, cmd, typename)
			local doc = mgr.current_document;
			if typename == nil then
				if doc.doc_type == doc_type.hex then
					print("HEX file");
				elseif doc.doc_type == doc_type.text then
					typename = content_type_name[doc.content_type];
					if typename then
						print("Current syntax type is:", typename);
					else
						print("Unknown text file(", doc.content_type, ")");
					end;
				end;
			else
				if typename == "off" then
					typename = "text";
				end;
				if typename ~= "?" then
					for type_id, name in pairs(content_type_name) do
						if name == typename then
							doc.content_type = type_id;
							print("Switch to ", typename, " mode.");
							return 1;
						end;
					end;
					print("Unknown syntax type: ", typename);
				end;
				local name_map = {};
				local name_array = {};
				for type_id, name in pairs(content_type_name) do
					if name_map[name] == nil then
						table.insert(name_array, name);
						name_map[name] = name;
					end;
				end;
				print("Available syntax types:", table.concat(name_array, "|"));
			end;
			return 1;
		end;
		["build"] = function(self, cmd, work)
			if work == nil then
				self:on_cmd(_op.build_target);
				return 1;
			end;
			if work == "project" then
				self:on_cmd(_op.build_project);
				return 1;
			end;
			return 0;
		end;
		["rebuild"] = function(self, cmd, work)
			if work == nil then
				self:on_cmd(_op.build_target);
				return 1;
			end;
			if work == "project" then
				self:on_cmd(_op.rebuild_project);
				return 1;
			end;
			return 0;
		end;
		["stop"] = function(self, cmd, work)
			if work == "build" then
				self:on_cmd(_op.stop_build);
				return 1;
			end;
			return 0;
		end;
		["log"] = function(self, cmd)
			local args = string.match(self.__raw_args, "^log%s+(.*)$");
			if args == nil or args == "" then
				print("log [file|command]");
				return 1;
			end;
			mgr:open_doc{3, args};
			return 1;
		end;
		["ssh"] = function(self, cmd, compress, host)
			if host == nil and (compress == nil or compress:sub(1,1) ~= "-") then
				host = compress;
				compress = nil;
			end;
			if compress == "-h" or host == nil then
				print("Usage: \n\tssh [-no-compress] ssh://<USR@>HOST<:PORT>");
				return 1;
			end;
			if compress == "-no-compress" then
				compress = " ";
			else
				compress = " -compress "
			end;
			local target = {host:match([[([^/]-://)([^/]+)(.*)]])};
			if target[1] == nil then
				local new_host = "ssh://"..host;
				target = {new_host:match([[([^/]-://)([^/]+)(.*)]])};
				if target[1] == nil then
					print("Unknown host: ", host);
					return 1;
				else
					host = new_host;
				end;
			end;
			if target[1]:lower() ~= "ssh://" then
				print("Unsupported host: ", host);
				return 1;
			end;
			local hostinfo = target[1]:lower() .. target[2];
			
			local sshcon = sshcon_path;
			local cmd = sshcon..compress..hostinfo.." \"$SHELL --login -i\"";
			edx:shell2(cmd, "*", "");
			return 1;
		end;
		["help"] = function(self, cmd, target)
			local NOTE_STYLE = "\x1b[33m";
			local NULL_STYLE = "\x1b[0m";
			local function help_cmd()
				print("\x1b[34;47mCommands\x1b[0m:");
				print("\tclear {target}");
				print("\tbuild [target]");
				print("\trebuild [target]");
				print("\tstop build");
				print("\tdebug [windbg|gdb] {target} [arguments] [working path]       ", NOTE_STYLE, "Start debugging.", NULL_STYLE);
				print("\tdebug map {local path} {remote path}                         ", NOTE_STYLE, "Mapping remote source path to local source path.", NULL_STYLE);
				print("\tgdb {target} [arguments] [working path]                      ", NOTE_STYLE, "Debugging with gdb.", NULL_STYLE);
				print("\twindbg {target} [arguments] [working path]                   ", NOTE_STYLE, "Debugging with windbg.", NULL_STYLE);
				print("\tlog [file|command]                                           ", NOTE_STYLE, "Open file/command in append mode.", NULL_STYLE);
				print("\treset lsp                                                    ", NOTE_STYLE, "Restart LSP service.", NULL_STYLE);
				print("\tsyntax {type}                                                ", NOTE_STYLE, "Change highlight type of the active document.", NULL_STYLE);
				print("\tssh ssh://host:port");
				print("\thelp [cmd|exp|func]");
			end;
			local function help_exp()
				print("\x1b[34;47mExpression eval\x1b[0m: {\x1b[33mEXPRESSION\x1b[0m},[\x1b[33mFORMAT\x1b[0m]");
				print("\t\x1b[33mEXPRESSION\x1b[0m: a valid c/c++ expression.");
				print("\t\tbuild in functions can be used.");
				print("\t\tdebug symbol is valid when debugging.");
				print("\t\x1b[33mFORMAT\x1b[0m: a format type.(\x1b[31mGDB symbols are not supported\x1b[0m)");
				print("\t\ts: string");
				print("\t\tws: wide string");
				print("\t\tu8: utf8 string");
				print("\t\tb: bin number");
				print("\t\tx: hex number");
				print("\t\to: oct number");
				print("\t\thr: HRESULT message");
				print("\t\twm: Win32 message");
				print("\t\tt: local time string");
				print("\t\tutc: UTC time string");
				print("");
				print("\texample1: 1 + 1 => \x1b[32m2\x1b[0m");
				print("\texample2: 1 > 2 => \x1b[32mfalse\x1b[0m");
				print("\texample3: (uint16_t)0x1234,b => \x1b[32m0010010'00110100\x1b[0m");
				print("\texample4: time(),t => current local time");
			end;
			local function help_func()
				print("\x1b[34;47mBuild-In functions\x1b[0m:");
				print("\ttime()\t\tcurrent gmt seconds");
				print("\tnan()\t\tdouble NaN");
				print("\tnanf()\t\tfloat NaN");
				print("\tabs(n)\t\tabsolute value of n");
				print("\tisnan(n)\ttest if n is a NaN value");
				print("\tround(n)\tround n");
				print("\tceil(n)\t\tround up n");
				print("\tfloor(n)\tround down n");
				print("\tsin(n)\t");
				print("\tasin(n)\t");
				print("\tcos(n)\t");
				print("\tacos(n)\t");
				print("\ttan(n)\t");
				print("\tatan(n)\t");
				print("\texp(n)\t");
				print("\texp2(n)\t");
				print("\tlog(n)\t");
				print("\tlog10(n)\t");
				print("\tlog2(n)\t");
				print("\tsqrt(n)\t");
				print("\tpow(n,p)\t");
				print("\tstrlen(str)\t");
				print("\tstrtod(str)\t");
				print("\tatoi(str)\t");
				print("\tbase64enc(str)\tencode the string with the base64 codec");
				print("\tbase64dec(str)\tdecode the base64 encoded string");
				print("\trand()\t\tgenerate a random number");
				print("\tsizeof(exp)\tcalculate size of the exp type");
			end;
			if target == "cmd" or target == "command" then
				help_cmd();
			elseif target == "exp" or target == "expression" then
				help_exp();
			elseif target == "func" or target == "function" then
				help_func();
			else
				help_cmd();
				print("");
				help_exp();
			end;
			return 1;
		end;
	};
};

function edx:on_handle_command(source, args)
	local handler = edx.command_handlers[source];
	handler = handler and handler[args[1]];
	if handler then
		self.__args = args;
		self.__raw_args = args["__raw"];
		return dbg_call(function() return handler(self, table.unpack(args)); end, "on_handler_command:");
	end;
	return 0;
end;

local function update_serial_ports()
	local ports = {utils:list_serial_ports()};
	table.sort(ports, function(a, b)
			local aidx = a[1]:lower():match("com(%d+)");
			local bidx = b[1]:lower():match("com(%d+)");
			if aidx and bidx then
				return tonumber(aidx) < tonumber(bidx);
			end;
			return a[1] < b[1];
	end);
	edx:update_serial_ports(ports);
end;

function edx:on_active_pane(pane)
	if pane == "serial_ports" then
		update_serial_ports();
	elseif pane == "class" then
		lsp:list_doc_structure(mgr.current_document, nil);
	end;
end;

update_serial_ports();

edx:update_serial_baudrate(SERIAL_BAUDRATES, SERIAL_DEFAULT_BAUDRATE);

function edx_do_build(message, do_build_params)
	if type(do_build_params) ~= "table" then
		print("Invalid param ", do_build_params);
		print("Type: ", type(do_build_params));
		return;
	end;
	mgr:clear_doc_diagnostic("compiler", "");
	local ret = edx.do_build(do_build_params);
	if ret then
		edx.clear("output");
		edx.clear("build_result");
		print(message);
		edx.active("output");
	elseif ret == nil then
		show_notify(LANG("notify/build/still_running"));
	else
		show_notify(LANG("notify/build/failed"));
	end;
end;

global_cmd_map = {
	[_op.xws_set_startup_target] = function(self, id)
		local item_path, item_type, item_id = xws_mgr.get_context_menu_item();
		if item_type ~= 0x11 then
			return;
		end;
		local item_name = xws_mgr.get_item_name(item_id);
		menu_bar.set_cmake_default_target(item_name);
		xws_mgr:update_targets({menu_bar.get_cmake_default_target()});
		cmake_update_include_path_for_opened_documents();
	end;
	[_op.xws_open_file_dir] = function(self, id)
		local item_path, item_type, item_id = xws_mgr.get_context_menu_item();
		if item_id == 0 then
			item_path = project:make_path() or "";
		end;
		local open_path = item_path or edx.startup_path;
		if item_type == 0x10 then
			local cmd = project:find_build_command(nil, true) or {build_path=""};
			open_path = cmd["build_path"];
		end;
		if item_type == 4 or item_type == 0x11 then
			-- file/target item
			if utils:file_time(item_path) then
				edx:shell_execute("open", "explorer.exe", [[/select,"]]..(item_path:gsub("[\"]", [[\%1]]))..[["]]);
			else
				-- item_path not exists
				open_path = make_path(item_path, "../");
				if utils:file_time(open_path) then
					-- parent path exists, open the path
					edx:shell_execute("open", make_path(open_path).."\\", "");
				end;
				show_notify(LANG("notify/file_not_found"):format(item_path));
			end;
		else
			edx:shell_execute("open", make_path(open_path).."\\", "");
		end;
	end;
	[_op.xws_open_in_new_indow] = function(self, id)
		local item_path, item_type, item_id = xws_mgr.get_context_menu_item();
		if item_id == 0 then
			item_path = project:make_path() or "";
		end;
		if item_type == 0x10 then
			local cmd = project:find_build_command(nil, true) or {build_path=""};
			item_path = cmd["build_path"];
		end;
		local cmd = string.format("\"%s\" -new-window \"%s\"", make_path(self.app_path, "edx.exe"), item_path);
		self:shell2(cmd, "*", self.startup_path);
	end;
	[_op.xws_open_ssh_terminal] = function(self, id)
		local item_path, item_type = xws_mgr.get_context_menu_item();
		local open_path = item_path or edx.startup_path;
		if item_type == 0x10 then
			local cmd = project:find_build_command(nil, true) or {build_path=""};
			open_path = cmd["build_path"];
		elseif item_type == 0x11 then
			local cmd = project:find_build_command(nil, true);
			local map_path = (cmd and cmd.toolset and (not cmd.toolset.wsl) and cmd.toolset.map_path) or function(path) return path; end;
			remote_item_path = map_path(item_path);
		end;
	end;
	[_op.xws_open_terminal] = function(self, id)
		local item_path, item_type, item_id = xws_mgr.get_context_menu_item();
		if item_id == 0 then
			item_path = project:make_path() or "";
		end;
		local open_path = item_path or edx.startup_path;
		if item_type == 0x10 then
			local cmd = project:find_build_command(nil, true) or {build_path=""};
			open_path = cmd["build_path"];
		end;
		if item_type == 4 or item_type == 0x11 then
			-- file/target item
			open_path = make_path(item_path, "../");
		end;
		if type(project.show_command_line) == "function" then
			project.show_command_line(open_path);
		else
			open_terminal(open_path,true);
		end;
	end;
	[_op.ext_git_open_gui] = function(self, id)
		local item_path, item_type = xws_mgr.get_selected_item();
		local open_path = item_path;
		if item_type == 4 then
			-- file item
			open_path = make_path(item_path, "../");
		end
		if open_path == "" then
			open_path = xws_mgr.workspace;
		end;
		edx.shell2([[git gui]], "", open_path);
	end;
	[_op.ext_git_open_history] = function(self, id)
		local item_path, item_type = xws_mgr.get_selected_item();
		local open_path = item_path;
		if item_type == 4 then
			-- file item
			open_path = make_path(item_path, "../");
		end
		if open_path == "" then
			open_path = xws_mgr.workspace;
		end;
		edx.shell2([[gitk]], "", open_path);
	end;
	[_op.ext_git_pull] = function(self, id)
		local item_path, item_type = xws_mgr.get_selected_item();
		local open_path = item_path;
		if item_type == 4 then
			-- file item
			open_path = make_path(item_path, "../");
		end
		if open_path == "" then
			open_path = xws_mgr.workspace;
		end;
		edx:clear("output");
		edx.shell2([[git pull --autostash --progress]], "output", open_path);
		edx:active("output");
	end;
	[_op.ext_git_push] = function(self, id)
		local item_path, item_type = xws_mgr.get_selected_item();
		local open_path = item_path;
		if item_type == 4 then
			-- file item
			open_path = make_path(item_path, "../");
		end
		if open_path == "" then
			open_path = xws_mgr.workspace;
		end;
		edx:clear("output");
		edx.shell2([[git push --progress]], "output", open_path);
		edx:active("output");
	end;
	[_op.xws_close_opend_file] = function(self,id)
		local item_path, item_type = xws_mgr.get_context_menu_item();
		if item_path == nil then
			return;
		end;
		if item_type == 4 then
			mgr.close_doc(item_path);
		else
			local all_docs = {mgr.get_all_documents()};
			local len = #item_path;
			for i, doc in ipairs(all_docs) do
				local doc_file = doc.get_file_name();
				if doc_file:sub(0,len) == item_path then
					mgr.close_doc(doc_file);
				end;
			end;
		end;
		mgr.current_document:redraw();
	end;
	[_op.xws_rename_selected_item] = function(self,id)
		xws_mgr:rename_selected_item();
	end;
	[_op.xws_new_file] = function(self, id)
		-- create file
		xws_mgr:create_item(4, true, true);
	end;
	[_op.xws_new_folder] = function(self, id)
		-- create folder
		xws_mgr:create_item(3, false, true);
	end;
	[RECENTLY_ITEM_ID_BASE] = function(self, id, flags)
		id = id - RECENTLY_ITEM_ID_BASE;
		local recently = edx.env.recently;
		local doc_name = recently[id];
		if doc_name == nil then
			doc_name = recently[tostring(id)];
		end;
		if utils:file_time(doc_name) == nil then
			self:remove_recently_doc(doc_name);
			show_notify(string.format(LANG("notify/file_not_found"), doc_name));
			return 1;
		end;
		if flags == 1 or (self:get_key_status("shift") and not self:get_key_status("alt") and not self:get_key_status("ctrl")) then
			-- open file in new window
			local cmd = string.format("\"%s\" -new-window \"%s\"", make_path(self.app_path, "edx.exe"), doc_name);
			self:shell2(cmd, "*", self.startup_path);
			return 1;
		end;
		mgr:open_doc(doc_name);
		return 1;
	end;
	[RECENTLY_FOLDER_ID_BASE] = function(self, id, flags)
		id = id - RECENTLY_FOLDER_ID_BASE;
		local recently = edx.env.recently_folders;
		local doc_name = recently[id];
		if doc_name == nil then
			doc_name = recently[tostring(id)];
		end;
		if utils:file_time(doc_name) == nil then
			self:remove_recently_folder(doc_name);
			show_notify(string.format(LANG("notify/folder_not_found"), doc_name));
			return 1;
		end;
		if flags == 1 or (self:get_key_status("shift") and not self:get_key_status("alt") and not self:get_key_status("ctrl")) then
			-- open dir in new window
			local cmd = string.format("\"%s\" \"%s\"", make_path(self.app_path, "edx.exe"), doc_name);
			self:shell2(cmd, "*", self.startup_path);
			return 1;
		end;
		lsp:terminate();
		self:stop_build();
		menu_bar:enable_cmake_menu_items();
		xws_mgr:load_xws(doc_name);
		return 1;
	end;
	-- CMAKE targets
	[CMAKE_TARGETS_ID_BASE] = function(self, id)
		menu_bar.set_cmake_default_target(id);
		xws_mgr:update_targets({menu_bar.get_cmake_default_target()});
		cmake_update_include_path_for_opened_documents();
		-- clean up
		return 1;
	end;
	-- CMAKE toolsets
	[CMAKE_TOOLSET_ID_BASE] = function(self, id)
		if menu_bar.set_cmake_toolset(id) then
			lsp:terminate();
			local toolset = menu_bar.get_cmake_toolset(true);
			project:on_toolset_changed(toolset);
			toolchain_update_include_path_for_opened_documents();
			if toolset["type"] == "STM32Cube" then
				-- STM32Cube project should rebuild cache when the toolset changed
				self:on_cmd(_op.cmake_rebuild_cache);
			else
				self:on_cmd(_op.cmake_build_cache);
			end;
		end;
		return 1;
	end;
	[CMAKE_SERIAL_PORTS_ID_BASE] = function(self, id)
		menu_bar.set_cmake_serial_port(id);
		return 1;
	end;
	[CMAKE_SERIAL_BUAD_ID_BASE] = function(self, id)
		menu_bar.set_cmake_serial_baud(id);
		return 1;
	end;
	[CMAKE_CPU_TYPES_ID_BASE] = function(self, id)
		menu_bar.set_cmake_cpu(id);
		menu_bar:disable_cmake_menu_items();
		lsp:terminate();
		esp32_idf_set_target(menu_bar.get_cmake_cpu(), menu_bar.get_cmake_config(), function(success)
				menu_bar:enable_cmake_menu_items();
				if success then
					local cmd = project:find_build_command(nil, true);
					lsp:reset(cmd);
					project:refresh_targets(cmd);
					cmake_update_include_path_for_opened_documents();
				end;
		end);
		return 1;
	end;
	[CMAKE_BOARD_TYPES_ID_BASE] = function(self, id)
		menu_bar.set_cmake_board(id);
		return 1;
	end;
	[CMAKE_DEBUG_ADAPTERS_ID_BASE] = function(self, id)
		if menu_bar.set_cmake_debug_adapter(id) then
			local toolset = menu_bar.get_cmake_toolset(true);
			local debug_adapter = toolset.debug_adapter;
			debug_adapter:switch_adapter(menu_bar.get_cmake_debug_adapter());
			menu_bar.update_cmake_debug_adapter_transport(debug_adapter.transports);
			menu_bar.update_cmake_debug_adapter_speed(debug_adapter.speeds);
			if #debug_adapter.transports == 0 then
				menu_bar.hide_cmake_items("debug_adapter_transport");
			else
				menu_bar.show_cmake_items("debug_adapter_transport");
			end;
			if #debug_adapter.speeds == 0 then
				menu_bar.hide_cmake_items("debug_adapter_speed");
			else
				menu_bar.show_cmake_items("debug_adapter_speed");
			end;
		end;
		return 1;
	end;
	[CMAKE_DEBUG_ADAPTER_SPEED_ID_BASE] = function(self, id)
		menu_bar.set_cmake_debug_adapter_speed(id);
		return 1;
	end;
	[CMAKE_DEBUG_ADAPTER_TRANSPORT_ID_BASE] = function(self, id)
		menu_bar.set_cmake_debug_adapter_transport(id);
		return 1;
	end;
	-- CMAKE config
	[CMAKE_CONFIG_ID_BASE] = function(self, id)
		if menu_bar.set_cmake_config(id) then
			self:on_cmd(_op.cmake_build_cache);
		end;
		return 1;
	end;
	-- CMAKE install
	[_op.cmake_install] = function(self, id)
		mgr:save_all_named();
		project:reload_settings();
		local cmd = project:find_build_command("install", true);
		if cmd ~= nil then
			cmd.is_install = true;
			local do_build_params = project:gen_do_build_params(cmd, "install", show_install_result_notification);
			__last_toolset = cmd.toolset;
			edx_do_build("Build & Install "..cmd["type"].." project", do_build_params);
			self.active("output");
		end
		return 1;
	end;
	[_op.cmake_build_cache] = function(self)
		mgr:save_all_named();
		project:reload_settings();
		local cmd = project:find_build_command(nil, true);
		if cmd ~= nil then
			local do_build_params = project:gen_do_build_params(cmd, nil, function(success)
					lsp:reset(cmd);
					cmake_update_include_path_for_opened_documents();
			end);

			__last_toolset = cmd.toolset;
			edx_do_build("Update "..cmd["type"].." cache", do_build_params);
		end
		self:active("output");
		return 1;
	end;
	[_op.cmake_rebuild_cache] = function(self)
		mgr:save_all_named();
		local cmd = project:find_build_command(".rebuild_cache", true);
		if cmd ~= nil then
			-- use local path for WSL target
			local map_path = (cmd.toolset and (not cmd.toolset.wsl) and cmd.toolset.map_path) or function(path) return path; end;
			local do_build_params = project:gen_do_build_params(cmd, nil, function(success)
					lsp:reset(cmd);
					cmake_update_include_path_for_opened_documents();
					if success then
						show_notify(LANG("notify/build/"..cmd["type"].."_cache_rebuilt"));
					else
						show_notify(LANG("notify/build/"..cmd["type"].."_cache_rebuild_failed"));
					end;
				end,
				{"@DELETE \"" .. map_path(cmd.build_path) .. "\""}
			);

			__last_toolset = cmd.toolset;
			edx_do_build("Rebuild "..cmd["type"].." cache", do_build_params);
			self.active("output");
		end
	end;
	[_op.cmake_refresh_cache] = function(self)
		mgr:save_all_named();
		local cmd = project:find_build_command(nil, true);
		if cmd ~= nil then
			local do_build_params = project:gen_do_build_params(cmd, nil, function(success)
					lsp:reset(cmd);
					cmake_update_include_path_for_opened_documents();
					if success then
						show_notify(LANG("notify/build/"..cmd["type"].."_cache_refreshed"));
					else
						show_notify(LANG("notify/build/"..cmd["type"].."_cache_refresh_failed"));
					end;
			end);

			__last_toolset = cmd.toolset;
			edx_do_build("Refresh "..cmd["type"].." cache", do_build_params);
			self.active("output");
		end
	end;
	-- 编译整个工程
	[_op.build_project] = function(self, id)
		mgr:save_all_named();
		project:reload_settings();
		local cmd = project:find_build_command();
		if cmd ~= nil then
			local do_build_params = project:gen_do_build_params(cmd, "build", show_build_result_notification);
			__last_toolset = cmd.toolset;
			edx_do_build("Build project", do_build_params);
			return 1;
		end
		self.clear("output");
		self.shell("mingw32-make all", "output");
		self.active("output");
		return 1;
	end; -- 编译单个目标
	[_op.build_target] = function(self, id, flags)
		mgr:save_all_named();
		project:reload_settings();
		local target_name = nil;
		local skip_current_doc = false;
		if flags == "from_xws_mgr" and self:is_focused("solution") then
			local item_path, item_type, item_id = xws_mgr.get_selected_item();
			if item_type == 0x11 then
				-- is the target item
				target_name = xws_mgr:get_item_name(item_id);
				skip_current_doc = true;
			end;
		end;
		local cmd = project:find_build_command(target_name, skip_current_doc);
		if cmd ~= nil then
			local do_build_params = project:gen_do_build_params(cmd, "build_target", show_build_result_notification);
			__last_toolset = cmd.toolset;
			edx_do_build("Build target", do_build_params);
			self.active("output");
			return 1;
		end;
		
		self.clear("output");
		self.shell("mingw32-make all", "output");
		self.active("output");
		return 1;
	end;
	[_op.xws_build_target] = function(self, id)
		return global_cmd_map[_op.build_target](self, id, "from_xws_mgr");
	end;
	[_op.xws_rebuild_target] = function(self, id)
		return global_cmd_map[_op.rebuild_target](self, id, "from_xws_mgr");
	end;
	[_op.build_and_run] = function(self, id)
		mgr:save_all_named();
		project:reload_settings();
		local target_name = nil;
		local skip_current_doc = false;
		if ESP32_IDF_CMAKE_PROJECT or BL_CMAKE_PROJECT or ZEPHYR_CMAKE_PROJECT then
			target_name = "flash";
		elseif self:is_focused("solution") then
			local item_path, item_type, item_id = xws_mgr.get_selected_item();
			if item_type == 0x11 then
				-- is the target item
				target_name = xws_mgr:get_item_name(item_id);
			end;
		end;
		local cmd = project:find_build_command(target_name, skip_current_doc);
		if cmd ~= nil then
			local on_complete = function(success)
				if success then
					local toolset = cmd.toolset;
					if ESP32_IDF_CMAKE_PROJECT or BL_CMAKE_PROJECT or ZEPHYR_CMAKE_PROJECT or toolset.debug_adapter then
						return;
					end;
					local wrap_command = toolset.wrap_command or function(cmd) return cmd; end;
					local map_path = toolset.map_path or function(path) return path; end;
					local file_type, file_type_name = edx:guess_file_type(cmd.target_output);
					local target_path = map_path(cmd.target_output);
					local cwd = (cmd.cwd or cmd.project_path);
					local target_cmd = target_path;
					local target_env = cmd.environment or {};
					if toolset.wsl or toolset.remote then
						target_cmd = target_cmd:gsub("([& \"%{%}%(%)])", [[\%1]]);
						if toolset.wsl then
							if cmd.argument then
								target_cmd = target_cmd.." "..cmd.argument;
							end;
							cwd = map_path(cwd);
							target_cmd = wrap_command(target_cmd, cwd);
							cwd = os.getenv("windir") or "c:\\windows";
						else
							if cmd.argument then
								target_cmd = target_cmd.." "..(cmd.argument:gsub("([\"])", [[\%1]]));
							end;
							target_cmd = wrap_command(target_cmd);
						end;
					else
						target_cmd = [["]]..target_cmd:gsub("([\"])", [[\%1]])..[["]];
						if cmd.argument then
							target_cmd = target_cmd.." "..cmd.argument;
						end;
					end;
					do
						-- add all target's path to PATH environment
						project:fix_path_env(target_env);
						cwd = table.clone(target_env, {[1] = cwd});
					end;

					if toolset.remote then
						local sshcon = sshcon_path;
						local cmd = sshcon.." -compress -pause "..toolset.remote.." \""..target_cmd.."\"";
						print(cmd);
						edx:shell2(cmd, "*", "");
					else
						if not (file_type_name and string.match(file_type_name, [[.*GUI EXE$]])) then
							-- for non-GUI executable target
							target_cmd = [[cmd /d /c ]] .. target_cmd .. [[ & pause]];
						end;
						edx:shell2(target_cmd, "*", cwd);
					end;
				else
					show_notify(LANG("notify/build/failed"));
				end;
			end;
			if is_debugger_running() then
				on_complete(true);
			else
				local do_build_params = project:gen_do_build_params(cmd, "build_and_run", on_complete);
				__last_toolset = cmd.toolset;
				edx_do_build("Build & Run target", do_build_params);
			end;
		else
			self.clear("output");
			self.shell("mingw32-make all -j2 & mingw32-make run", "output");
			self.active("output");
		end;
		return 1;
	end;
	[_op.rebuild_project] = function(self, id)
		mgr:save_all_named();
		project:reload_settings();
		local cmd = project:find_build_command();
		if cmd ~= nil then
			local do_build_params = project:gen_do_build_params(cmd, "rebuild", show_build_result_notification);
			__last_toolset = cmd.toolset;
			edx_do_build("Rebuild "..cmd["type"].." project", do_build_params);
			return 1;
		end;

		self.clear("output");
		self.shell("mingw32-make rebuild -j2", "output");
		self.active("output");
		return 1;
	end;
	[_op.rebuild_target] = function(self, id, flags)
		mgr:save_all_named();
		project:reload_settings();
		local target_name = nil;
		local skip_current_doc = false;
		if flags == "from_xws_mgr" and self:is_focused("solution") then
			local item_path, item_type, item_id = xws_mgr.get_selected_item();
			if item_type == 0x11 then
				-- is the target item
				target_name = xws_mgr:get_item_name(item_id);
			end;
		end;
		local cmd = project:find_build_command(target_name, skip_current_doc);
		if cmd ~= nil then
			local do_build_params = project:gen_do_build_params(cmd, "rebuild_target", show_build_result_notification);
			__last_toolset = cmd.toolset;
			edx_do_build("Rebuild "..cmd["type"].." target", do_build_params);
			return 1;
		end;

		self.clear("output");
		self.shell("mingw32-make rebuild -j2", "output");
		self.active("output");
		return 1;
	end;
	[_op.stop_build] = function(self, id)
		print("Stop build");
		self:stop_build();
		menu_bar:enable_cmake_menu_items();
	end;
	
	[_op.search_everywhere] = function(self, id)
		print("search_everywhere");
		edx:search_everywhere();
	end;

	-- 列举所有断点
	[_op.debug_list_breakpoint] = function(self, id)
		list_breakpoints();
		return 0;
	end;
	[_op.active_toolsets] = function(self, id)
		toolset_manager:show();
		return 0;
	end;
	[_op.register_shell_menu_shortcuts] = function(self, id)
		edx:register_shell_menu();
	end;

	[_op.open_config_file] = function(self,id)
		mgr.open_doc(edx_config_path);
	end;

	[_op.locate_current_document] = function(self,id)
		if xws_mgr.is_loaded then
			local cur_doc = mgr.current_document;
			xws_mgr:select_item(cur_doc.file_name);
			edx:active("solution");
		else
			show_notify(LANG("notify/project_not_loaded"));
		end;
	end;

	[_op.save_all] = function(self,id)
		mgr:save_all(0);
	end;

	[_op.open_homepage] = function(self, id)
		edx:shell_execute("open", "https://www.ed-x.cc", "");
	end;

	[_op.zoom_in] = function(self, id)
		local new_scalar = mgr.scalar * 10;
		if new_scalar >= 20 then
			new_scalar = new_scalar + 2;
		else
			new_scalar = new_scalar + 1;
		end;
		if new_scalar >= 30 then
			new_scalar = 30;
		end;
		new_scalar = math.ceil(math.floor(new_scalar * 10) / 10) / 10;
		mgr.scalar = new_scalar;
	end;
	[_op.zoom_out] = function(self, id)
		local new_scalar = mgr.scalar*10;
		if new_scalar > 20 then
			new_scalar = new_scalar - 2;
		else
			new_scalar = new_scalar - 1;
		end;
		if new_scalar < 5 then
			new_scalar = 5;
		end;
		new_scalar = math.ceil(math.floor(new_scalar * 10) / 10) / 10;
		mgr.scalar = new_scalar;
	end;
	[_op.zoom_reset] = function(self, id)
		mgr.scalar = 1.0;
	end;
	[_op.zoom_50] = function(self, id)
		local _zoom_value_map = {
			[_op.zoom_50] = 0.5;
			[_op.zoom_80] = 0.80;
			[_op.zoom_120] = 1.2;
			[_op.zoom_150] = 1.5;
			[_op.zoom_200] = 2.0;
			[_op.zoom_250] = 2.5;
			[_op.zoom_300] = 3.0;
		};
		mgr.scalar = _zoom_value_map[id];
	end;
	[_op.zoom_80] = function(self, id)
		global_cmd_map[_op.zoom_50](self, id);
	end;
	[_op.zoom_120] = function(self, id)
		global_cmd_map[_op.zoom_50](self, id);
	end;
	[_op.zoom_150] = function(self, id)
		global_cmd_map[_op.zoom_50](self, id);
	end;
	[_op.zoom_200] = function(self, id)
		global_cmd_map[_op.zoom_50](self, id);
	end;
	[_op.zoom_250] = function(self, id)
		global_cmd_map[_op.zoom_50](self, id);
	end;
	[_op.zoom_300] = function(self, id)
		global_cmd_map[_op.zoom_50](self, id);
	end;
	[_op.serial_monitor] = function(self, id)
		esp32_idf_show_serial_monitor();
	end;

	[_op.new_project] = function(self, id)
		self:create_project();
	end;

	[_op.delay_refresh_class_pane] = function(self, id)
		if edx:is_visible("class") then
			lsp:list_doc_structure(mgr.current_document, nil);
		end;
	end;

	[_op.new_edx_instance] = function(self, id)
		local cmd = string.format("\"%s\" -new-window", make_path(self.app_path, "edx.exe"));
		self:shell2(cmd, "*", self.startup_path);
		return 1;
	end;

	[_op.reopen_last_file] = function(self, id)
		edx:reload_recently_menu();
		if not edx.env then
			return;
		end;
		for idx, path in ipairs(edx.env.recently or {}) do
			local doc = mgr:get_document(path);
			if not doc then
				mgr:open_doc(path);
				break;
			end;
		end;
	end;
};

function edx:on_cmd(id, flags)
	return dbg_call(
		function()
			local item;
			local op_id = id;
			if id >= _op.active_pane_first and id <= _op.active_pane_last then
				-- active/hide panes
				item = global_cmd_map[id];
				if item then
					item(self, op_id, flags);
				else
					edx:active_pane(op_id);
				end;
				return 0;
			end;
			if id >= 0xE000 and id <= 0xFFFF then
				id = id - (id % 0x20);
			end;
			item = global_cmd_map[id];
			-- fallback to previous command handler
			while id >= 0xE000 and item == nil do
				id = id - 0x20;
				item = global_cmd_map[id];
			end;
			if item ~= nil then
				return item(self, op_id, flags);
			end;
			return 0;
		end
	);
end;

function edx:on_fullscreen(fullscreen)
	self.is_fullscreen = fullscreen;
	if xws_mgr.is_loaded then
		if fullscreen then
			self:switch_layout("fullscreen");
		else
			self:switch_layout("simple");
		end;
	else
		if fullscreen then
			self:switch_layout("fullscreen_lite");
		else
			self:switch_layout("lite");
		end;
	end;
end;

function edx:msdn_search(keyword, devlang)
	local search_sites = {
		["C++"] = [[+(site%3acppreference.com+OR+site%3amicrosoft.com+OR+site%3aman7.org)]];
		["javascript"] = [[+(site%3amozilla.org+OR+site%3aw3.org)]];
		["lua"] = [[+(site%3alua.org)]];
		["cmake"] = [[+(site%3acmake.org)]];
		["__default__"] = [[]];
	};
	local text = ([["https://bing.com/search?q=%s%s&ensearch=1"]]):format(keyword:gsub(" ", "%20"), search_sites[devlang or "__default__"]);
	self:shell_execute("open", text, "");
end;

text_cmd_map = {
	[_op.help] = function(self, id)
		local doc = mgr.current_document;
		local devlang = nil;
		if doc.content_type == content_type_cxx then
			-- C/C++文档
			devlang = "C++";
		elseif doc.content_type == content_type_js then
			-- JavaScript文档
			devlang = "javascript";
		elseif doc.content_type == content_type_cmake then
			-- JavaScript文档
			devlang = "cmake";
		elseif doc.content_type == content_type_lua then
			devlang = "lua";
		else
			return 0;
		end ;
		local selection = { doc.get_selection() };
		if selection[1] == nil or selection[1] ~= selection[3] then
			edx.do_cmd(_op.select_word);
			selection = { doc.get_selection() };
		end ;
		if selection[1] == nil then
			return 0;
		end ;
		local text = doc.get_text(selection[1], selection[2], selection[4] - selection[2]);
		edx:msdn_search(text, devlang);
	end;
	[_op.help_search] = function(self, id)
		local doc = mgr.current_document;
		local devlang = nil;
		if doc.content_type == content_type_cxx then
			-- C/C++文档
			devlang = "C++";
		elseif doc.content_type == content_type_js then
			-- JavaScript文档
			devlang = "javascript";
		else
			return 0;
		end ;
		local selection = { doc.get_selection() };
		if selection[1] == nil or selection[1] ~= selection[3] then
			edx.do_cmd(_op.select_word);
			selection = { doc.get_selection() };
		end ;
		if selection[1] == nil then
			return 0;
		end ;
		local text = doc.get_text(selection[1], selection[2], selection[4] - selection[2]);
		edx:msdn_search(text, devlang);
	end;
	[_op.paste] = function(self, id)
		-- 如果粘贴的行数小于100行,则自动格式化这些行
		local fl = self.current_document.cursor_line;
		local fp = self.current_document.cursor_column;
		edx.do_cmd(_op.paste_ex);
		local el = self.current_document.cursor_line;
		local ep = self.current_document.cursor_column;
		if math.abs(el - fl) < 100 then
			edx.do_cmd(_op.format_selection);
		else
			edx.do_cmd(_op.unselect);
		end ;
		return 1;
	end;
	[_op.indent_split_line_or_jump_placeholder] = function(self, id)
		if lsp:jump_placeholder(self.current_document) then
			edx.do_cmd(_op.update_completion_tip);
			return 1;
		end;
		edx:do_cmd(_op.indent_split_line);
		return 1;
	end;
	[_op.unselect_or_hide_search_box] = function(self, id)
		local doc = self.current_document;
		if doc.is_selection_valid then
			edx:do_cmd(_op.unselect);
			return 1;
		end;
		self.show_search_box(false);
		return 1;
	end;
	[0x01003] = function(self, id)
		local doc = mgr.current_document;
		local sel_text = doc.selection_text;
		if sel_text ~= nil then
			local new_text = sel_text:upper();
			if new_text ~= sel_text then
				doc.replace(new_text, true);
			end ;
		end ;
		return 1;
	end;
	[0x01004] = function(self, id)
		local doc = mgr.current_document;
		local sel_text = doc.selection_text;
		if sel_text ~= nil then
			local new_text = sel_text:lower();
			if new_text ~= sel_text then
				doc.replace(new_text, true);
			end ;
		end ;
		return 1;
	end;
	[0x01005] = function(self, id)
		local doc = mgr.current_document;
		if doc.content_type ~= content_type_cxx then
			-- 不是C++类型
			return 0;
		end ;
		return 1;
	end;
	-- 查找引用
	[_op.find_reference] = function(self, id)
		local doc = mgr.current_document;
		if lsp:find_reference(doc) then
			return 1;
		end;
		return 1;
	end;
	[_op.dump_bin] = function(self, id)
		local doc = mgr.current_document;
		local doc_name = doc.file_name;
		local doc_ext_name = doc_name:match([[.+%.(%w+)]]):lower();
		if doc.content_type == 0 or doc.doc_type == doc_type.hex then
			if  (
					doc_ext_name == 'dll' or
					doc_ext_name == 'ocx' or
					doc_ext_name == 'exe' or
					doc_ext_name == 'lib' or
					doc_ext_name == 'obj' or
					doc_ext_name == 'a' or
					doc_ext_name == 'o'
				)
			then
				local cmd = ([[%s\edx-tools\dumpbin.exe /NOLOGO /SUMMARY /HEADERS /SYMBOLS /EXPORTS /IMPORTS /ARCHIVEMEMBERS /DEPENDENTS "%s"]]):format(edx.app_path, doc_name);
				edx.shell(cmd, "new:text");
			elseif doc_ext_name == 'apk' then
				local cmd = ([[pushd %s\edx-tools && aapt.exe d badging "%s" && echo -------- && echo File Content: && aapt l -a "%s"]]):format(edx.app_path, doc_name, doc_name);
				edx.shell(cmd, "new:text");
			end ;
			return 0;
		end ;
		return 1;
	end;
	-- 重复选择内容
	[0x01007] = function(self, id)
		local doc = mgr.current_document;
		local sel_text = doc.selection_text;
		local fl, fp, el, ep = doc.get_selection();
		if sel_text ~= nil then
			doc.set_cursor(el, ep);
			doc.set_text(sel_text);
			doc.select_text(el, ep, doc.cursor_line, doc.cursor_column);
			doc.reset_render();
		else
			fl = doc.cursor_line;
			fp = doc.cursor_column;
			doc.select_text(fl, 0, fl, -1);
			sel_text = doc.selection_text;
			doc.set_cursor(fl, 0);
			doc.set_text(sel_text .. "\n");
			doc.set_cursor(fl + 1, fp);
		end ;
	end;
	-- 查找所有实现
	[_op.find_implementation] = function(self, id)
		local doc = mgr.current_document;
		lsp:find_implementation(doc);
		return 0;
	end;
	-- 加载当前文件为系统脚本/C++列出文件结构
	[_op.list_doc_structure] = function(self, id)
		local doc = mgr.current_document;
		if doc.content_type == content_type_cxx then
			-- c/c++ file
			lsp:list_doc_structure(doc, mgr.popup_class_view);
			return;
		end;
		local file_name = doc.file_name;
		local run_func = function()
			dofile(utils:convert_string(file_name, "utf8", "acp"));
		end;
		if file_name == nil or #file_name == 0 then
			local line_index = 0;
			local internal_name = doc.internal_name or "<NONAME>"
			local func, err_info = load(
				function()
					local line;
					line = doc:get_text(line_index, 0, 4096);
					line_index = line_index + 1;
					if line ~= nil then
						line = line.."\r\n";
					end;
					return line;
				end, internal_name
			);
			if func == nil then
				print("can not load lua script! ");
				err_info = err_info:gsub("([ \t]*)%[string \"(doc://<%x*>)\"]:(%d+):","%1%2:%3:");
				print(err_info);
				edx:active("output");
				return;
			else
				doc.content_type = content_type_lua;
			end
			run_func = func;
		end
		xpcall(run_func,
			function(msg)
				self.clear("output");
				print("error:\r\n");
				msg = debug.traceback(msg);
				msg = msg:gsub("([ \t]*)%[string \"(doc://<%x*>)\"]:(%d+):","%1%2:%3:");
				msg = msg:gsub("\n", "\r\n");
				print(msg);
				print("\r\n");
				edx:active("output");
				return 0;
			end
		);
	end;
	-- hover info
	[0x0100A] = function(self, id)
		local doc = mgr.current_document;
		lsp:hover_info(doc, nil, nil, true);
		return 1;
	end;
	[_op.jump_to_definition] = function(self, id)
		local doc = mgr.current_document;
		lsp:find_definition(doc);
		return 1;
	end;
	[_op.jump_to_declaration] = function(self, id)
		local doc = mgr.current_document;
		lsp:find_declaration(doc);
		return 1;
	end;
	[_op.close_other_files] = function(self, id)
		local cur_doc = mgr.current_document;
		local all_docs = {mgr.get_all_documents()};
		for i, doc in ipairs(all_docs) do
			if cur_doc ~= doc then
				mgr.close_doc(doc);
			end;
		end;
		cur_doc:redraw();
	end;
	[_op.open_file_dir] = function(self, id)
		local cur_doc = mgr.current_document;
		edx.shell_execute("open", "explorer.exe", [[/select,"]]..cur_doc.file_name..[["]]);
	end;
	[_op.open_file_terminal] = function(self, id)
		local cur_doc = mgr.current_document;
		local open_path = make_path(cur_doc.file_name, "..");
		open_terminal(open_path);
	end;
	-- change format
	[CMAKE_FMT_ID_BASE] = function(self, id)
		id = id - CMAKE_FMT_ID_BASE;
		local is_reload = (id < 0x20);
		local format = id % 0x20;
		local doc = mgr.current_document;
		doc.change_format(is_reload, format);
	end;
	[SYNTAX_MENU_ID_BASE] = function(self, id)
		id = id - SYNTAX_MENU_ID_BASE;
		local doc = mgr.current_document;
		local doc_type_id = doc.doc_type;
		if doc_type_id == doc_type.text and id == content_type_md_visual then
			mgr.convert_doc_type(doc, doc_type.markdown);
			return 1;
		end;
		if doc_type_id == doc_type.markdown then
			mgr.convert_doc_type(doc, doc_type.text);
			if id < content_type_last then
				doc = mgr.current_document;
				doc.content_type = id;
			end;
			return 1;
		end;
		if id == content_type_md then
			if doc_type_id == doc_type.markdown then
				mgr.convert_doc_type(doc, doc_type.text);
			elseif doc_type_id == doc_type.text then
				doc.content_type = id;
			end;
			return 1;
		elseif id < content_type_last and doc_type_id == doc_type.text then
			doc.content_type = id;
		end;
		return 1;
	end;
	[LINE_FORMAT_MENU_ID_BASE] = function(self, id)
		id = id - LINE_FORMAT_MENU_ID_BASE;
		local doc = mgr.current_document;
		doc.line_format = id;
		return 1;
	end;
	[_op.open_tab_dir] = function(self, id)
		if edx.__last_tab_doc then
			edx.shell_execute("open", "explorer.exe", [[/select,"]]..edx.__last_tab_doc.name..[["]]);
			edx.__last_tab_doc = nil;
		end;
	end;
	[_op.open_tab_terminal] = function(self, id)
		if edx.__last_tab_doc then
			local open_path = make_path(edx.__last_tab_doc.name, "..");
			open_terminal(open_path);
			edx.__last_tab_doc = nil;
		end;
	end;
	[_op.copy_tab_path] = function(self, id)
		if edx.__last_tab_doc then
			mgr.set_clipboard(edx.__last_tab_doc.name);
			edx.__last_tab_doc = nil;
		end;
	end;
	[_op.close_tab] = function(self, id)
		if edx.__last_tab_doc then
			mgr.close_doc(edx.__last_tab_doc.id);
			mgr.current_document.redraw();
			edx.__last_tab_doc = nil;
		end;
	end;
	[_op.close_other_tabs] = function(self, id)
		if edx.__last_tab_doc then
			local cur_doc_id = edx.__last_tab_doc.id

			local all_docs = {mgr.get_all_documents()};
			for i, doc in ipairs(all_docs) do
				if cur_doc_id ~= doc.id then
					mgr.close_doc(doc);
				end;
			end;
			local cur_doc = mgr.current_document;
			cur_doc:redraw();
		end;
	end;
	[_op.locate_tab_file] = function(self, id)
		if edx.__last_tab_doc then
			xws_mgr:select_item(edx.__last_tab_doc.name);
			edx:active("solution");
			edx.__last_tab_doc = nil;
		end;
	end;
	["document_idle_handlers"] = {
		[content_type_cxx] = function(self, id, doc)
			edx:post_command_delay(1000, _op.delay_refresh_class_pane, 12345);
			local last_cmd = self.__last_cmd[1];
			local accept_cmds = {
				[_op.smart_write_left_parenthesis] = 1;
				[_op.smart_write_left_square_bracket] = 1;
			};
			if not (type(last_cmd) == "string" and last_cmd:match("[#_a-zA-Z]")) and not accept_cmds[last_cmd] then
				return;
			end;
			if self.__completing ~= nil then
				return;
			end;
			-- is typing
			if not doc.should_do_completion then
				return;
			end;
			if doc.cursor_column < 256 then
				local text = doc.get_text(doc.cursor_line, 0, doc.cursor_column);
				if text ~= nil then
					if text:match("([^ \t]+)") == nil then
						return;
					end;
					local match_text = text:match("^%s*#%s*include%s+[\"<].*$");
					if match_text == nil and doc.cursor_line < 50 then
						match_text = text:match("^%s*//.*$");
					end;
					if match_text == nil then
						match_text = text:match("^%s*#%s*%w*$");
					end;
					if match_text ~= nil then
						-- include filename
						-- preprocessor
						-- instant compile instructions
						edx:do_cmd(_op.show_completion_list);
						return;
					end;
				end;
			end;
			lsp:do_completion(doc);
		end;
		[content_type_lua] = function(self, id, doc)
			-- edx:post_command_delay(1000, _op.delay_refresh_class_pane, 12345);
			local last_cmd = self.__last_cmd[1];
			if not (type(last_cmd) == "string" and last_cmd:match("[a-z]")) then
				return;
			end;
			if self.__completing ~= nil then
				return;
			end;
			-- is typing
			if doc.cursor_column < 256 then
				local text = doc.get_text(doc.cursor_line, 0, doc.cursor_column);
				if text ~= nil then
					local word = text:match("^[ \t]+(%w+)");
					if word == nil then
						return;
					end;
					if word == "end" or
						word == "else" or
						word == "elseif" then
						edx:do_cmd(_op.format_selection);
						return;
					end;
				end;
			end;
		end;
	};
	[_op.document_idle] = function(self, id)
		table.remove(self.__last_cmd, 1); -- strip document_idle command
		local doc = self.current_document;
		local handler = text_cmd_map["document_idle_handlers"][doc.content_type];
		if handler then
			handler(self, id, doc);
		end;
	end;
	[_op.smart_write_left_parenthesis] = function(self, id)
		local doc = self.current_document;
		if doc.content_type == content_type_cxx then
			if self.__completing then
				edx:do_cmd(id);
				lsp:do_completion(doc);
				return 1;
			end;
		end;
		return 0;
	end;
	[_op.smart_write_right_parenthesis] = function(self, id)
		local doc = self.current_document;
		if doc.content_type == content_type_cxx then
			if self.__completing then
				edx:do_cmd(id);
				lsp:do_completion(doc);
				return 1;
			end;
		end;
		return 0;
	end;
	[_op.smart_write_left_square_bracket] = function(self, id)
		local doc = self.current_document;
		if doc.content_type == content_type_cxx then
			if self.__completing then
				edx:do_cmd(id);
				lsp:do_completion(doc);
				return 1;
			end;
		end;
		return 0;
	end;
	[_op.smart_write_right_square_bracket] = function(self, id)
		local doc = self.current_document;
		if doc.content_type == content_type_cxx then
			if self.__completing then
				edx:do_cmd(id);
				lsp:do_completion(doc);
				return 1;
			end;
		end;
		return 0;
	end;
	[_op.smart_write_left_bracket] = function(self, id)
		local doc = self.current_document;
		if doc.content_type == content_type_cxx then
			if self.__completing then
				edx:do_cmd(id);
				lsp:cancel_completion(doc);
				edx.do_cmd(_op.completion_end_tip);
				edx.do_cmd(_op.completion_end_list);
				return 1;
			end;
		end;
		return 0;
	end;
	[_op.smart_write_right_bracket] = function(self, id)
		local doc = self.current_document;
		if doc.content_type == content_type_cxx then
			if self.__completing then
				edx:do_cmd(id);
				lsp:do_completion(doc);
				return 1;
			end;
		end;
		return 0;
	end;
	["smart_del_op_handler"] = function(self, id)
		local doc = self.current_document;
		if doc.content_type == content_type_cxx then
			if self.__completing then
				edx:do_cmd(id);
				if doc.cursor_column < 256 then
					local text = doc.get_text(doc.cursor_line, 0, doc.cursor_column);
					if text ~= nil and text:match("([^ \t]+)") == nil then
						return 1;
					end;
				end;
				lsp:do_completion(doc);
				return 1;
			end;
		end;
		return 0;
	end;
	[_op.smart_backspace] = function(self, id)
		return self:smart_del_op_handler(id);
	end;
	[_op.smart_delete] = function(self, id)
		return self:smart_del_op_handler(id);
	end;
	[_op.smart_write_quote] = function(self, id)
		if self.__completing then
			return 0;
		end;
		edx:do_cmd(id);
		local doc = self.current_document;
		if doc.content_type == content_type_cxx then
			local text = doc.get_text(doc.cursor_line, 0, doc.cursor_column);
			if text ~= nil then
				local match_text = text:match("^[ \t]*#[ \t]*include[ \t]+[\"<]$");
				if match_text ~= nil then
					-- 自动补全include
					edx.do_cmd(_op.show_completion_list);
				end;
			end;
		end;
		return 1;
	end;
	[_op.smart_fix] = function(self, id)
		local doc = self.current_document;
		lsp:find_code_action(doc, "quickfix");
		return 1;
	end;
	[_op.scroll_up] = function(self, id)
		local doc = self.current_document;
		doc:scroll(false, -2);
		return 1;
	end;
	[_op.scroll_down] = function(self, id)
		local doc = self.current_document;
		doc:scroll(false, 2);
		return 1;
	end;
};

function mgr:on_cmd(id, flags)
	while #self.__last_cmd > 10 do
        table.remove(self.__last_cmd);
	end;
    table.insert(self.__last_cmd, 1, id);
	local op_id = id;
    local item;
    if id >= 0xE000 and id <= 0xFFFF then
		id = id - (id % 0x20);
	end;
	item = text_cmd_map[id];
	-- fallback to previous command handler
	while id >= 0xE000 and item == nil do
		id = id - 0x20;
		item = text_cmd_map[id];
	end;
    if item ~= nil then
		return item(self, op_id, flags);
    end ;
    return 0;
end;

function edx:on_locate_output(info)
	local file_name;
	local out;
	if info:sub(1,7):lower() == "http://"
		or info:sub(1,8):lower() == "https://"
		or info:sub(1,6):lower() == "cmd://"
		or info:sub(1,6):lower() == "ssh://"
	then
		-- for http, https, cmd, ssh
		out = { info:match("^%s*(%w+://.+):(%d+),(%d+):(%d+),(%d+):") };
		if out[1] == nil or out[2] == nil then
			return 0;
		end;
		file_name = out[1];
	else
		if project:is_xmake() then
			-- strip terminal highlight info
			info = info:gsub("\x1b%[.-m", "");
			local real_info = info:match("^error:%s+(.*)$");
			if real_info == nil then
				real_info = info:match("^warning:%s+(.*)$");
			end;
			if real_info then
				info = real_info;
			end;
		end;
		-- match for vc
		out = { info:match("^%s*([a-zA-Z]*:?[^<>?*:;'\"]+)%((%d+)%)%s*:") };
		if out[1] == nil or out[2] == nil then
			out = { info:match("^%s*(doc://<[%d%w]+>):(%d+),(%d+):(%d+),(%d+):") };
		else
			-- vc output has no column info
			out[3] = 1;
		end;
		if out[1] == nil or out[2] == nil then
			out = { info:match("^%s*(doc://<[%d%w]+>):(%d+):") };
			out[3] = 1;
		end;
		if out[1] == nil or out[2] == nil then
			-- hex doc
			out = { info:match("^%s*(doc://<[%d%w]+>):(%d+)-(%d+):") };
			out[3] = 1;
		end;
		if out[1] == nil or out[2] == nil then
			-- hex doc
			out = { info:match("^%s*([a-zA-Z]?:?[^<>?*:;'\"]+):(%d+)-(%d+):") };
		end;
		if out[1] == nil or out[2] == nil then
			-- clang tips
			out = { info:match("^%s*In file included from ([a-zA-Z]?:?[^<>?*:;'\"]+):(%d+)[:,]") };
			out[3] = 1;
		end;
		if out[1] == nil or out[2] == nil then
			-- gcc tips
			out = { info:match("^%s*inlined from.*%s+at%s+([a-zA-Z]?:?[^<>?*:;'\"]+):(%d+):(%d+)[:,]") };
		end;
		if out[1] == nil or out[2] == nil then
			out = { info:match("^%s*from ([a-zA-Z]?:?[^<>?*:;'\"]+):(%d+)[:,]") };
			out[3] = 1;
		end;
		if out[1] == nil or out[2] == nil then
			-- clang
			out = { info:match("^%s*([a-zA-Z]?:?[^<>?*:;'\"]+):(%d+):(%d+):") };
		end;
		if out[1] == nil or out[2] == nil then
			-- clang-cl
			out = { info:match("^%s*([a-zA-Z]?:?[^<>?*:;'\"]+)%((%d+),(%d+)%):") };
		end;
		if out[1] == nil or out[2] == nil then
			out = { info:match("^%s*([a-zA-Z]?:?[^<>?*:;'\"]+):(%d+):") };
			out[3] = 1;
		end;
		if out[1] == nil or out[2] == nil then
			out = { info:match("^([a-zA-Z]?:?[^<>?*:;'\"]+):%d+%((%d+)%)") };
			out[3] = 1;
		end;
		if out[1] == nil or out[2] == nil then
			out = { info:match("^%s*([a-zA-Z]?:?[^<>?*:;'\"]+):(%d+),(%d+):(%d+),(%d+):") };
		end;
		if out[1] == nil or out[2] == nil then
			out = { info:match("^%s*(\\\\wsl%$\\[^<>?*:;'\"]+):(%d+),(%d+):(%d+),(%d+):") };
		end;
		if out[1] == nil or out[2] == nil then
			-- for keil C51 compiler
			out = { info:match("^%s*[*]+.*IN LINE (%d+) OF ([a-zA-Z]?:?[^<>?*:;'\"]+):") };
			out[1], out[2], out[3] = out[2], out[1], 1
		end;
		if out[1] == nil or out[2] == nil then
			-- cmake error
			out = { info:match("^%s*CMake Error at%s+([a-zA-Z]?:?[^<>?*:;'\"]+):(%d+)") };
			out[3] = 1;
		end;
		if out[1] == nil or out[2] == nil then
			-- cmake error include
			out = { info:match("^%s+([a-zA-Z]?:?[^<>?*:;'\"]+):(%d+)%s+%(include%)") };
			out[3] = 1;
		end;

		if out[1] == nil or out[2] == nil then
			return 0;
		end;
		file_name = out[1];
		if file_name:sub(1,7):lower() ~= "doc://<" then
			local regular_file_name = project:regular_path(file_name);
			if regular_file_name and regular_file_name ~= "" then
				file_name = regular_file_name;
			end;
			if file_name:match([[^\\wsl%$\.*$]]) then
				-- keep the wsl path
			elseif __last_toolset and __last_toolset.unmap_path then
				local unmap_path = __last_toolset.unmap_path;
				file_name = unmap_path(file_name);
			end;

			if project:is_ready() and utils:file_time(file_name) == nil and file_name:sub(1,7):lower() ~= "doc://<" then
				file_name = project:make_path("dummy-build-path", file_name);
			end;
		end;
	end;

	if mgr.open_doc(file_name) then
		mgr.record_cursor_history();
		local cur_doc = mgr.current_document;
		if cur_doc.doc_type == doc_type.hex then
			cur_doc:select_hex(tonumber(out[2]), tonumber(out[3]));
		else
			cur_doc:set_cursor(tonumber(out[2]) - 1, tonumber(out[3]) - 1);
			if out[4] ~= nil and out[5] ~= nil then
				cur_doc:select_text(tonumber(out[2]) - 1, tonumber(out[3]) - 1, tonumber(out[4]) - 1, tonumber(out[5]) - 1);
			end ;
		end;
        edx.do_cmd(_op.center_cursor_line);
    end;
end;

dbg_call(
        function()
			update_short_cuts(_op_scope.text, { _op.toggle_comment_selection, k("ctrl+/") });

			-- update_short_cuts( _op_scope._global, { _op.debug_start, k("f5") } );
			-- update_short_cuts( _op_scope._global, { _op.debug_step_over, k("f10") } );
			-- update_short_cuts( _op_scope._global, { _op.debug_step_in, k("f11") } );
			-- update_short_cuts( _op_scope._global, { _op.debug_step_out, k("shift+f11") } );
			-- update_short_cuts( _op_scope._global, { _op.debug_terminate, k("shift+f5") } );
			-- update_short_cuts( _op_scope._global, { _op.debug_toggle_breakpoint, k("f9") } );
			-- update_short_cuts( _op_scope._global, { _op.debug_break, k("ctrl+f9") } );
			-- update_short_cuts( _op_scope._global, { _op.debug_jump_to, k("alt+f9") } );
			update_short_cuts( _op_scope._global, { _op.locate_current_document, k("alt+l") } );


			update_short_cuts(_op_scope.text, { 0x01003, k("ctrl+shift+u") });    -- toupper
            update_short_cuts(_op_scope.text, { 0x01004, k("ctrl+u") });    -- tolower
            update_short_cuts(_op_scope.text, { 0x01005, k("ctrl+k"), k("ctrl+i") });    -- show C/C++ context info
            update_short_cuts(_op_scope.text, { 0x01007, k("ctrl+d") }); -- duplicate selection
			update_short_cuts(_op_scope.text, { 0x0100A, k("ctrl+q") });    -- hover info

            -- 加载EDX环境
            edx:load_env();
			-- 切换到默认的布局

			if xws_mgr.is_loaded then
				edx:switch_layout("simple");
			else
				edx:switch_layout("lite");
			end;

			make_edx_completion();
        end,
		LANG("notify/error/init_edx_failed")
);

-- 加载debug绑定
require "dbg"

function list_toolsets()
	if TOOLSETS == nil or #TOOLSETS == 0 then
		print("no toolset was found!");
		return;
	end;
	print("Found C/C++ toolset(s)!");
	for i, toolset in ipairs(TOOLSETS) do
		print("\t", i, ":", toolset.name, " => ", toolset.home);
	end ;
end;

function update_toolsets_menu()
	menu_bar.update_cmake_toolset(TOOLSETS, nil, function(tool)
			local name = tool.name;
			local vers = nil;
			local sdks = nil;
			local active_ver = nil;
			local active_sdk = nil;
			if (tool.vers and #tool.vers > 1) or (tool.sdks and #tool.sdks > 1) then
				vers = tool.vers;
				sdks = tool.sdks;
				active_ver = tool.version;
				active_sdk = tool.sdk;
				name = tool.name.."|"..tool.version.."|"..tool.sdk;
			end;
			return name, name.."\t"..tool.home, active_ver, vers, active_sdk, sdks, function(ver, sdk)
				if ver == nil then
					ver = tool.version;
				end;
				if sdk == nil then
					sdk = tool.sdk;
				end;
				tool:switch_version(ver, sdk);
			end;
	end);
end;

function update_toolsets()
	search_compilers(function(toolsets)
		TOOLSETS = toolsets;
		if toolsets ~= nil then
			list_toolsets();
			edx:load_toolset();
			update_toolsets_menu();
		end;
	end);
end;

dbg_call(update_toolsets, LANG("notify/error/init_toolset_failed"));

require "project_wizard"
require "configuration"

-- 加载启动脚本
-- {USER}/AppData/Local/user-config.lua
-- {EDX HOME}/user-config.lua
for idx, path in ipairs(user_config_path) do
	if path and utils:file_time(path) then
		pinvoke(dofile, path);
		break;
	end;
end;
